r"""PAM-wraper GUI wrapper for the AVG Linux GUI

Copyright (C) GRISOFT, Inc.
"""

try:
	from sys import exit
	import signal
	import os
	from stat import *
	import sys
	import sre
	import pwd
	import pygtk
	pygtk.require('2.0')
	import gtk
	import gobject
	import logger
except Exception, detail:
	print "** CRIT: %s", detail
	exit(1)

class CPAMwrap:
	"""
	CPAMwrap is class for wrapping PAM-wrapper with GUI.
	
	Arguments:
	path -- absolute path to PAM-wrapper
	
	Attributes:
		
	Methods:
	get_exit_status() -- get the exit status of PAM-wrapped program
	get_usr_data() -- get user-defined data from PAM-wrapper		
	run() -- run PAM-wrapper (authenticate and run wrapped program (blocking))
	
	As a PAM-wrapper can be used for example userhelper from package usermode. 
	""" 
	

	def __init__(self, path):

		# codes of protocol prompt-type in pipe communication with PAM-wrapper	
		self.__pw_code = {
			'PW_SYNC_POINT' 		: '32',		# synchronization (shouldn't speak both in the same time) 

			# PAM-wrapper request codes (text):
			'PW_UNKNOWN_PROMPT'		: '33',	# PAM-wrapper can't figure out PAM request 
			'PW_ECHO_ON_PROMPT'		: '34',	# PAM request - PAM_PROMPT_ECHO_ON PAM message
			'PW_ECHO_OFF_PROMPT'	: '35',	# PAM request - PAM_PROMPT_ECHO_OFF PAM message
			'PW_PROMPT_SUGGESTION'	: '36',	# suggestion of response to message
			'PW_INFO_MSG' 			: '37',	# PAM_TEXT_INFO PAM message
			'PW_ERROR_MSG' 			: '38',	# PAM_ERROR_MSG PAM message
			'PW_EXPECT_RESP'		: '39',	# number of responses expected by PAM-wrapper
			'PW_SERVICE_NAME'		: '40',	# service required
			'PW_FALLBACK_ALLOW' 	: '41',	# 0/1 fallback allowed ?
			'PW_USER'				: '42',	# user required to use service
			'PW_BANNER'				: '43',	# text to be displayed in dialog
			'PW_EXEC_START' 		: '44',	# announcement of executing PAM-wrapped command
			'PW_EXEC_FAILED'		: '45',	# announcement of failing of executed PAM-wrapped command
			
			'PW_USR_DATA'			: '66', # user defined data
		}
	
		# CPAMwrap response codes (bytes):
		self.__resp_code = {
			'PW_SYNC_POINT' : '\040',   # 32
			'PW_TEXT'		: '\041',	# 33
			'PW_CANCEL'		: '\042',	# 34
			'PW_FALLBACK'	: '\043',	# 35
		}

		# disctionary mapping of
		# Valid PAM-wrapper exit status to CPAMwrapEroor exception name
		self.__pw_status_name = {
			# exit codes of PAM-wrapper (compatibile with userhelper, pamwrap)
			1: 'PASSWD_INVALID', # password is not right
			2: 'FIELDS_INVALID', # gecos fields invalid or sum(lengths) too big 
			3: 'SET_PASSWORD', # password resetting error 
			4: 'LOCKS',	# some files are locked 
			5: 'NO_USER',	# user unknown ... 
			6: 'NO_RIGHTS',	# insufficient rights  
			7: 'INVALID_CALL',	# invalid call to PAM-wrapper 
			8: 'SHELL_INVALID',	# 
			9: 'NO_MEMORY',	# out of memory 
			10: 'NO_PROGRAM',	# progname (PAM-wrapped program) not found 
			11: 'EXEC_FAILED',	# exec failed for some reason 
			12: 'CANCELED',	# user cancelled operation 
			13: 'PAM_INT_ERROR',	# PAM internal error
			20: 'USAGE' # invalid usage of PAM-wrapper 
			}
		
		self.__logger = logger.CLogger()

		# check if the path of PAM-wrapper is secure

		# path to PAM-wrapper must be absolute		
		if (not os.path.isabs(path)):
			raise ValueError, 'given path %s is not absolute' % path

		# checking PAM-wrapper permissions (owner is root, suid is set, ...)
		# ? TODO: not so restrictive permissions?	
		if (os.stat(path).st_uid == 0
			and os.access(path, os.X_OK)
			and oct(S_IMODE(os.stat(path).st_mode)) == '04711'):	
			pass
		else:
			raise IOError, '%s has bad permissions or owner' % path

		# check if PAM is available
		if (not os.access('/etc/pam.d/avggui', os.F_OK)):
			raise CPAMwrapError('NO_PAM')

		self.__pw_path = path

		# PAM-wrapper is runnng
		self.__pw_is_running = False
	
		# Used PAM-wrapper has to communicate with CPAMwrap via file-descriptors 4 (out) and 3 (in):		
		self.__pw_in_fd = 3
		self.__pw_out_fd = 4

		# pipe protocol:
		# pattern of prompts from PAM-wrapper (lines with prompt-type text code (number) and prompt (string))			
		self.__prompt_pattern = '\s*(\d+) *([^\n]*)\n(.*)$'
		self.__prompt_re = sre.compile(self.__prompt_pattern, sre.DOTALL)

		# conversation dialog padding
		self.__table_pad = 8
		# response_id for fallback button
		self.__response_fallback = 100

				
	def run(self, parent_win, argv='', banner='', run_info=None, can_kill = False):
		"""
		Start the PAM-wrapper.
				
		arguments:
		parent_win -- Parent window of PAM-conversation dialog.
		argv -- Arguments for PAM-wrapper.
		        (Typically contains wrappped program to run and his arguments.)
		banner -- Text to be displayed at the top of conversation dialog.
		          Overrides the banner given by PAM-wrapper.
		run_info -- Information to be displayed in blocking waiting-dialog
		            while wrapped program is running.
		can_kill -- flag - wrapped program can be killed by user from waiting-dialog
		
		returns: 
		- True if PAM-wrapper exited with 0.
		- In other cases (e.g. failed authentication or error) raises exception CPAMwrapError 
		"""
		
		self.__pw_argv = [self.__pw_path] + argv
		self.__parent_win = parent_win
		self.__banner = banner
		self.__run_info = run_info
		self.__can_kill = can_kill
		self.__waiting_dialog = None;
				
		if self.__pw_is_running:
			raise CPAMwrapError('WRP_RUNNING')
		self.__pw_is_running = True
		
		# exit status of PAM-wrapper
		self.__pw_status = None
		# exit status of wrapped program
		self.__child_exitstatus = None
		# user-defined data (typically stdout and/or stderr of program wrapped by PAM-wrapper):
		self.__usr_data = ''
		# user data from PAM-wrapper are being read 
		self.__reading_usr_data = False

		# PAM-wrapped program was execed by PAM-wrapper
		self.__child_was_execed = False
		# PAM-wrapper exited
		self.__pw_exited = False

		# PAM-wrapper PID
		self.__pw_pid = None		

		# init structure holding response to PAM-wrapper
		self.__init_response()

		# use two pipes for comunicaton with PAM-wrapper
		try:	
			childout_r, childout_w = os.pipe()
			childin_r, childin_w = os.pipe()
		except OSError, detail:
			self.__logger.err(detail)
			return

		try:
			self.__pw_pid = os.fork()
		except OSError, detail:
			self.__logger.err(detail)
			return
			
		if not self.__pw_pid:
			# child
			try:
				os.close(childin_w)
				os.close(childout_r)	
			except OSError, detail:
				self.__logger.err(detail)
				return
		
			# synchronization with parent
			os.read(childin_r, 2)
			
			try:							
								
				# PAM-wrapper (will be execed) expects fd 3 for reading and fd 4 for writing
	
				# First create two copies of stdin, in case 3 and 4 aren't
				# currently in use.
	
				dummy_fd0 = os.dup(sys.stdin.fileno());
				dummy_fd1 = os.dup(sys.stdin.fileno());
	
				# Now create temporary copies of the pipe descriptors, which
				# aren't int to be 3 or 4 because they are surely in use
				# by now.
		
				tmp_childin_r = os.dup(childin_r);
				tmp_childout_w = os.dup(childout_w);
	
				# Now get rid of the temporary descriptors,
				os.close(dummy_fd0);
				os.close(dummy_fd1);
				os.close(childin_r);
				os.close(childout_w);
	
				# and move the pipe descriptors to their now homes.
				os.dup2(tmp_childin_r, self.__pw_in_fd)
				os.dup2(tmp_childout_w, self.__pw_out_fd)	
	
				os.close(tmp_childin_r)
				os.close(tmp_childout_w)
			
				self.__logger.info('running PAM-wrapper %s with args %s' % (self.__pw_path, self.__pw_argv))
				os.execv(self.__pw_path, self.__pw_argv)		 

			except OSError, detail:
				print detail
				self.__logger.err(detail)	
				return	
		else:
			# parent
			
 			# close the write-end of the reading pipe,
			# and the read-end of the writing pipe.
			os.close(childout_w);
			os.close(childin_r);

			self.__childout_r = childout_r
			# use file instead of file descriptor to be able to flush
			self.__childin_w_file = os.fdopen(childin_w, 'w', 0)

			saved_handler = signal.signal(signal.SIGCHLD, self.__pw_exited_cb)
			gobject.io_add_watch(childout_r, gobject.IO_IN, self.__read_childout_cb)
			gobject.io_add_watch(childout_r, gobject.IO_ERR, self.__read_childout_err)
			gobject.io_add_watch(childout_r, gobject.IO_HUP, self.__read_childout_hup)	

			self.__logger.debug("sending \"Go\" to PAM-wrapper (start authentication)")
			# Tell the child we're ready for it to run.
			os.write(childin_w, "Go")

			# active blocking waiting for PAM-wrapper exit
			gtk.main()
			#signal.signal(signal.SIGCHLD, signal.SIG_DFL)
			signal.signal(signal.SIGCHLD, saved_handler)
			if self.__exc_name:
				raise CPAMwrapError(self.__exc_name)
			return self.__run_return
			

	def __parse_exitstatus(self, status):
		if self.__child_was_execed:
			self.__child_exitstatus = status
			self.__exc_name = 'PROGRAM_EXIT'
		elif (status in self.__pw_status_name.keys()):
			self.__exc_name = self.__pw_status_name[status]
		else:
			self.__exc_name = 'UNK_ERROR'
		
	def __pw_exited_cb(self, foo, data):
		# callback on PAM-wrapper exit
		self.__logger.debug("PAM-wrapper exited")
		self.__pw_exited = True
		self.__stop()
		return False
	
	def __stop(self):
		# stop running PAM-wrapper
		self.__logger.debug("__stop() called; exited=%s reading=%s" % (self.__pw_exited, self.__reading_usr_data))	
		if (self.__pw_exited and not self.__reading_usr_data and self.__pw_is_running):
			self.__pw_is_running = False

			# destroy waiting dialog if needed
			if self.__waiting_dialog != None:
				self.__waiting_dialog.destroy()
							   
			pid, status = os.waitpid(self.__pw_pid, 0)

			# this is also the case when running wrapped program is canceled from dialog 
			if os.WIFSIGNALED(status):
				self.__run_return = False
				self.__exc_name = 'PW_SIGNAL_EXIT'

			elif os.WIFEXITED(status):
				self.__pw_status = os.WEXITSTATUS(status)
				self.__logger.debug("Exit status: %d" % self.__pw_status)
				if self.__pw_status:
					self.__run_return = False
					self.__parse_exitstatus(self.__pw_status)
				else:
					self.__run_return = True
					self.__exc_name = ''					

			else:
				self.__run_return = False
				self.__exc_name = 'UNK_ERROR'					

			gtk.main_quit()
									

	def __read_childout_cb(self, event, data):

		if self.__reading_usr_data:
			# reading user data (PW_USER_DATA protocol prompt-type) 
			# (should only occur at the end of communication with PAM-wrapper,
			#  so all resting input will be read as user-data)
			try:
				pw_output = os.read(self.__childout_r, os.sysconf('SC_LINE_MAX'))
			except OSError, detail:
				self.__logger.err(detail)
				raise 
			if pw_output == '':
				self.__logger.debug("All user data read")	
				self.__reading_usr_data = False	
				self.__stop()
				return False # remove callback
			else:
				self.__logger.debug('Reading user data %s' % pw_output)
				# TODO: refactor (use list, strings aren't effective in Python) 
				self.__usr_data = self.__usr_data + pw_output
				return True	 
			
		# reading prompts from PAM-wraper
		self.__logger.debug('Reading childout')
		try:
			pw_output = os.read(self.__childout_r, os.sysconf('SC_LINE_MAX'))
		except OSError, detail:
			self.__logger.err(detail)
			raise
			
		if pw_output == '': 			 # EOF
			self.__logger.debug("Callback canceled")
			return False	# remove callback
		else:
			self.__parse_pw_output(pw_output)
		return True

	def __parse_pw_output(self, outline):

		self.__logger.debug('Parsing PAM-wrapper output:\n %s' % outline)	
		
		while outline:
		
			try:
				prompt_type, prompt, new_outline = self.__prompt_re.match(outline).groups()
			except AttributeError:
				raise CPAMwrapError('PROTOCOL')

			outline = new_outline	

			# structure to hold a message from PAM-wrapper (PAM-prompt)	
			msg = {	
				'type'		: prompt_type,
				'message'	: prompt,
				'label'		: None, # conversation dialog label of prompt
				'entry'		: None  # conversation dialog entry of prompt
			}	
			
			echo = True # echo entered data

			if (prompt_type == self.__pw_code['PW_USR_DATA'] or self.__reading_usr_data):
				# user data come at the end of all data from PAM-wrapper (last prompt of protocol)
				self.__reading_usr_data = True
				self.__logger.debug('Reading user data %s' % outline)
				# TODO: refactor
				self.__usr_data = self.__usr_data + outline
				return
			elif prompt_type == self.__pw_code['PW_PROMPT_SUGGESTION']:
				self.__resp['suggestion'] = prompt
		
			elif (prompt_type == self.__pw_code['PW_ECHO_OFF_PROMPT']
			      or prompt_type == self.__pw_code['PW_ECHO_ON_PROMPT']):
				if prompt_type == self.__pw_code['PW_ECHO_OFF_PROMPT']:
					echo = False
					
				if not self.__resp['title']:
					self.__resp['title'] = 'Query'				
			
				label = gtk.Label('_'+prompt)
				label.set_use_underline(True)
				label.set_line_wrap(True)			
				label.set_alignment(1.0, 0.5)
				
				entry = gtk.Entry()
				entry.set_visibility(echo)
				label.set_mnemonic_widget(entry)			
	
				msg['label'] = label
				msg['entry'] = entry
				
				if self.__resp['suggestion']:
					entry.set_text(self.__resp['suggestion'])
					self.__resp['suggestion'] = None				
				
				if not self.__resp['first']:
					self.__resp['first'] = entry
				self.__resp['last'] = entry
				
				self.__resp['table'].attach(label, 0, 1, self.__resp['rows'], self.__resp['rows']+1, gtk.EXPAND | gtk.FILL, 0, self.__table_pad, self.__table_pad)
				self.__resp['table'].attach(entry, 1, 2, self.__resp['rows'], self.__resp['rows']+1, gtk.EXPAND | gtk.FILL, 0, self.__table_pad, self.__table_pad)
	
				self.__resp['message_list'].append(msg)
				self.__resp['responses'] = self.__resp['responses'] + 1
				self.__resp['rows'] = self.__resp['rows'] + 1
	
			elif prompt_type == self.__pw_code['PW_FALLBACK_ALLOW']:
				if int(prompt) != 0:
					self.__resp['fallback_allowed'] = True
				else:
					self.__resp['fallback_allowed'] = False
	
			elif prompt_type == self.__pw_code['PW_USER']:
				if ((prompt.find('<user>') == -1) and (prompt.find('<none>') == -1)):
					self.__resp['user'] = prompt 
	
			elif prompt_type == self.__pw_code['PW_SERVICE_NAME']:
				self.__resp['service'] = prompt 
				
			elif prompt_type == self.__pw_code['PW_ERROR_MSG']:
				self.__resp['title'] = "Error" 
				msg['label'] = gtk.Label(prompt)
				self.__resp['table'].attach(msg['label'], 0, 2, self.__resp['rows'], self.__resp['rows']+1, 0, 0, self.__table_pad, self.__table_pad)	
				self.__resp['message_list'].append(msg)
				self.__resp['rows'] = self.__resp['rows'] + 1
				
			elif prompt_type == self.__pw_code['PW_INFO_MSG']:									
				self.__resp['title'] = "Information"
				msg['label'] = gtk.Label(prompt)
				self.__resp['table'].attach(msg['label'], 0, 2, self.__resp['rows'], self.__resp['rows']+1, 0, 0, self.__table_pad, self.__table_pad)	
				self.__resp['message_list'].append(msg)
				self.__resp['rows'] = self.__resp['rows'] + 1
						
			elif prompt_type == self.__pw_code['PW_BANNER']:									
				self.__resp['banner'] = prompt 			
			
			elif prompt_type == self.__pw_code['PW_EXEC_START']:									
				self.__child_was_execed = True
				
			elif prompt_type == self.__pw_code['PW_EXEC_FAILED']:									
				self.__child_was_execed = False
			
			elif prompt_type == self.__pw_code['PW_EXPECT_RESP']:															
				if self.__resp['responses'] != int(prompt):
					raise CPAMwrapError('PROTOCOL')
			
			elif prompt_type == self.__pw_code['PW_SYNC_POINT']:															
			 	self.__resp['ready'] = True
		
		if (self.__resp['ready'] and self.__resp['responses'] == 0):
			self.__write_pw(gtk.RESPONSE_OK)
			self.__init_response()
		elif (self.__resp['ready'] and self.__resp['responses'] > 0):

			# PAM conversation dialog
			dialog = gtk.Dialog('', self.__parent_win, gtk.DIALOG_MODAL)
			self.__resp['dialog'] = dialog
			
			if self.__resp['responses'] > 0:
				dialog.add_buttons(gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
			else:
				dialog.add_buttons((gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
			
			if self.__resp['title']:
				dialog.set_title(self.__resp['title'])
			else:
				dialog.set_title('Error')
								
			dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
			pixbuf = gtk.gdk.pixbuf_new_from_file("pixmaps/password.png")
			dialog.set_icon(pixbuf)

			info_hbox = gtk.HBox()
			dialog.vbox.add(info_hbox)
			dialog.vbox.add(self.__resp['table'])

			if self.__resp['responses'] > 0:
				pixbuf = gtk.gdk.pixbuf_new_from_file("pixmaps/keyring.png")
				image = gtk.Image()
				image.set_from_pixbuf(pixbuf)
				info_hbox.add(image)
			
			dialog.connect('map-event', self.__grab_keyboard_cb)			
			dialog.connect('delete-event', self.__conv_dialog_delete_event_cb)
			
			if self.__resp['responses'] == 0:
				text = ''
			elif self.__resp['service'] == 'passwd':
				text = '' 
				if self.__resp['service'] == 'chfn':
					text = 'Changing personal information.'
				elif self.__resp['service'] == 'chsh':
					text = 'Changing login shell'
				elif self.__resp['banner']:
					text = self.__resp['banner']
				elif self.__resp['fallback_allowed']:
					text = 'You are attempting to run %s which may benefit from administrative privileges, but more information is needed in order to do so' % self.__resp['service']								
				else:												
					text = 'You are attempting to run %s which requires administrative privileges, but more information is needed in order to do so' % self.__resp['service']											
			else:
				if self.__resp['banner']:
					text = self.__resp['banner']
				else:					
					if self.__resp['fallback_allowed']:
						text = 'You are attempting to run %s which may benefit from administrative privileges, but more information is needed in order to do so' % self.__resp['service']								
					else:												
						text = 'You are attempting to run %s which requires administrative privileges, but more information is needed in order to do so' % self.__resp['service']											
		
			label = gtk.Label(text)
			label.set_line_wrap(True)
			info_hbox.add(label)
			
			if (self.__resp['fallback_allowed'] and self.__resp['responses'] > 0):
				dialog.add_button('Run Unprivileged', self.__response_fallback)
			if self.__resp['last']:
				self.__resp['last'].connect('activate', self.__fake_respond_ok_cb, dialog)
	
			#dialog.set_size_request(400, -1)

			# method run() of gtk.Dialog can't be used because when it is used
			# with gdk threads support it freezes
			self.__logger.debug("Running conversation dialog")	
			self.__run_conv_dialog(dialog)

			if self.__resp['first']:
				self.__resp['first'].grab_focus

			gtk.gdk.keyboard_ungrab(0L)
			
			self.__init_response()
						
		if self.__child_was_execed:
			# after child (wrapped program) was execed (we know it from just parsed data),
			# run a dialog blocking till the end of child execution 
			if self.__can_kill: # child can be killed by user
				# Cancel-dialog	
				w_dialog = gtk.MessageDialog(self.__parent_win, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_CANCEL, self.__run_info)
				w_dialog.connect('response', self.__waiting_dialog_response_cb)
			else:
				# Info-dialog
				w_dialog = gtk.MessageDialog(self.__parent_win, gtk.DIALOG_MODAL, gtk.MESSAGE_INFO, gtk.BUTTONS_NONE, self.__run_info)
			self.__waiting_dialog = w_dialog	
			self.__logger.debug("Running waiting dialog")	
			w_dialog.connect('delete-event', self.__waiting_delete_event_cb)
			w_dialog.connect('close', self.__waiting_delete_event_cb)
			w_dialog.show_all()		

	def __run_conv_dialog(self, dialog):
		dialog.connect('response', self.__conv_dialog_response_cb)
		dialog.show_all()		

		gtk.gdk.threads_enter()
		gtk.main()
		gtk.gdk.threads_leave()

	def __conv_dialog_response_cb(self, dialog, response_id):
		if response_id == gtk.RESPONSE_DELETE_EVENT:
			#dialog.emit_stop_by_name('delete_event')
			return True # stops signal propagation to default callback
		gtk.main_quit()
		self.__write_pw(response_id, dialog)

	def __waiting_delete_event_cb(self, dialog, data):
		return True # stops signal propagation to default callback
		
	def __waiting_dialog_response_cb(self, dialog, response_id):
		if response_id == gtk.RESPONSE_DELETE_EVENT:
			#dialog.emit_stop_by_name('delete_event')
			return True # stops signal propagation to default callback
		self.__waiting_dialog.destroy()
		if self.__can_kill:
			self.__kill_pw()

	def __kill_pw(self):
		self.__logger.debug("Killing pw")	
		os.kill(self.__pw_pid, signal.SIGKILL)
		
	def __write_pw(self, response, dialog=None):
		# write responses (stored in conversation dialog) to PAM-wrapper
		# arguments:
		#    response: -- gtk response_id from dialog box
		
		if response == self.__response_fallback:
			self.__childin_w_file.write(self.__resp_code['PW_FALLBACK']+'\n')
		elif response == gtk.RESPONSE_CANCEL:	
			self.__childin_w_file.write(self.__resp_code['PW_CANCEL']+'\n')
			# flush()

		elif response == gtk.RESPONSE_OK:
			for msg in self.__resp['message_list']:
				if msg['entry']:
					self.__childin_w_file.write(self.__resp_code['PW_TEXT']+msg['entry'].get_text()+'\n')
					# flush()
			self.__childin_w_file.write(self.__resp_code['PW_SYNC_POINT']+'\n')			

		else:
			raise CPAMwrapError('UNK_ERROR')
			# We were closed, deleted, canceled, or something else
			# which we can treat as a cancellation.

		if dialog:
			dialog.destroy() # conversation dialog can be destroyed only after getting responses from dialog entries 
				
	def __init_response(self):
		# stores response to PAM-wrapper
		self.__resp = {
			'responses'			: 0,		# number of responses required; int
			'rows'				: 0,		# number of rows in dialog; int
			'ready'				: False,	# ready to run dialog; True/False
			'fallback_allowed'	: False,	# ; True/False
			'user'				: pwd.getpwuid(os.getuid())[0],		# ; string
			'service'			: '',		# PAM-service; string
			'suggestion'		: '',		# response suggestion; string
			'banner'			: self.__banner,		# banner from PAM-wrapper to display in dialog; string
			'title'				: '',		# dialog title; string
			'message_list'		: [],		# list of received messages; list of tuples (prompt type, prompt, widget label, widget entry)
			'dialog'			: None,		# dialog widget (messagebox); gtk widget
			'first'				: None,		# first entry of dialog; gtk widget
			'last'				: None,		# last entry of dialog; gtk widget
			'table'				: gtk.Table(2, 1, False),		# input part of dialog; gtk widget
		}
		
	def __fake_respond_ok_cb(self, entry, dialog):
		dialog.response(gtk.RESPONSE_OK)
		
	def __grab_keyboard_cb(self, dialog, data):
		pass

	def __read_childout_err(self, fd, data):
		self.__logger.debug('CPAMwrap got signal ERR on reading pipe')
		raise CPAMwrapError('UNK_ERROR')
		
	def __read_childout_hup(self, fd, data):
		self.__logger.debug('CPAMwrap got signal HUP on reading pipe')
		self.__reading_usr_data = False
		self.__stop()
		
	def __conv_dialog_delete_event_cb(self, dialog, data):
		return True # stops signal propagation to default callback
	
	def get_usr_data(self):
		# Get user data form PAM-wrapper (e.g. stdout and/or stderr of wrapped program)
		if self.__reading_usr_data:
			raise CPAMwrapError('READING_DATA', 'can\'t get user data, PAM-wrapper is reading data')
		if self.__pw_is_running:
			raise CPAMwrapError('WRP_RUNNING', 'can\'t get user data, PAM-wrapper is running')
		return self.__usr_data

	def get_exit_status(self):
		# Get exit status of wrapped program 
		if self.__pw_is_running:
			raise CPAMwrapError('WRP_RUNNING', 'can\'t get exit status, PAM-wrapper is running')
		return self.__child_exitstatus
	
		
class CPAMwrapError:
	"""
	CPAMwrapError is exception class for method run() of CPAMwrap.

	Arguments:
	name -- symbolic name of exception
	msg -- text message of exception
	
	Attributes:
	name -- symbolic name of exception
	code -- code of exception
	msg -- text message of exception

	Returns:
	(name, code, msg)

	"""

	def __init__(self, name, msg=''):

		self.__err_table = {
			'UNK_EXIT':(-1, 'unknown exit code'),
			'OK':(0, 'information updated'),
			# exit codes of PAM-wrapper (e.g. userhelper, pamwrap)
			'PASSWD_INVALID':(1, 'invalid password'), # password is not right
			'FIELDS_INVALID':(2, 'invalid gecos field'), # gecos fields invalid or sum(lengths) too big 
			'SET_PASSWORD':(3, 'password resetting error'), # password resetting error 
			'LOCKS':(4, 'some system files are locked'),	# some files are locked 
			'NO_USER':(5, 'unknown user'),	# user unknown ... 
			'NO_RIGHTS':(6, 'insufficient rights'),	# insufficient rights  
			'INVALID_CALL':(7, 'invalid call to PAM-wrapper'),	# invalid call to PAM-wrapper 
			'SHELL_INVALID':(8, 'current shell not in /etc/shells'),	#
			'NO_MEMORY':(9, 'out of memory'),	# out of memory 
			'NO_PROGRAM':(10, 'invalid call to PAM-wrapped program'),	# progname (PAM-wrapped program) not found 
			'EXEC_FAILED':(11, 'exec() call failed'),	# exec failed for some reason 
			'CANCELED':(12, 'operation canceled by user'),	# user cancelled operation 
			'PAM_INT_ERROR':(13, 'internal PAM error'),	# PAM internal error
			'USAGE':(20, 'invalid PAM-wrapper usage'),	# invalid PAM-wrapper usage
			# end of exit codes of PAM-wrapper
			'PROGRAM_EXIT':(120, 'subprocess exited with error'), # means that authentication was passed
			'WRP_RUNNING':(121, 'PAM-wrapper running'),
			'READING_DATA':(122, 'data are being read from PAM-wrapper'),
			'PROTOCOL':(123, 'PAM-wrapper protocol error'),
			'NO_PAM':(124, 'can\'t use PAM module'),
			'PW_SIGNAL_EXIT':(254, 'PAM-wrapper exited due to received signal'),
			'UNK_ERROR':(255, 'unknown error')	# unknown error				
			}


		self.name = name
		self.code = self.__err_table[name][0] 
		if msg:
			self.msg = msg
		else:
			self.msg = self.__err_table[name][1] 
		self.__args = (self.code, self.name, self.msg)
		
	def __str__(self):
		return self.msg		

	def __getitem__(self, key):
		return self.__args[key]
	

	
################################################################
# testing code

if __name__ == "__main__":

	window = gtk.Window(gtk.WINDOW_TOPLEVEL)

	# run() method arguments:
	parent_window = window
	args_to_pamwrap = ['-w', '/home/vykydal/workspace/korendel/testuid', '4', '0', '1>/dev/null'] # wrapped program and his arguments
	#args_to_pamwrap = ['-o', '3', '-w', '/home/vykydal/workspace/korendel/testuid', '4', '0'] # get stderr and stdout of wrapped prg
	#args_to_pamwrap = [] # only authenticate
	conversation_dialog_banner = 'PAM authentication'
	waiting_dialog_text = 'Wait, wrapped program is running ...'
	can_kill_wrapped_prg = True
	
	def run_cb(widget):
		try:
			result = w.run(parent_window, args_to_pamwrap, conversation_dialog_banner, waiting_dialog_text, can_kill_wrapped_prg)
			print "run() method result: %s" % result

		except CPAMwrapError, inst:
			code, name, msg = inst
			print 'Exception raised:\n  code: %d\n  name: %s\n  message: %s' % (code, name, msg)
			if name == 'PROGRAM_EXIT':
				print "Wrapped program exitcode: %d" % w.get_exit_status()
		else:
			print "Output data of wrapped prg (stdout + stderr):\n %s" % w.get_usr_data()

			
	def quit_cb(widget):
		gtk.main_quit()


#	w = CPAMwrap('/usr/sbin/userhelper')   # should work OK with userhelper from package usermode
	w = CPAMwrap('/home/vykydal/workspace/korendel/rpamwrap')
	
	vbox = gtk.VBox()
	window.add(vbox)
	button = gtk.Button("Run test")
	button.connect("clicked", run_cb)
	vbox.add(button)
	button = gtk.Button("Quit")
	button.connect("clicked", quit_cb)
	vbox.add(button)
	window.set_position(gtk.WIN_POS_CENTER)
 	window.show_all()
	# test proper working with gdk threads support
	gtk.gdk.threads_init()
	gtk.main()

