"""
Module for holding configuration of AVG GUI for Linux
It provides abstract class CAvgGuiConfig for holding configuration values 
and define set of opertions for manipulating them.

Copyright (C) GRISOFT, s.r.o.
"""

import ConfigParser
import inspect
import types
import sre

from os import write
from shutil import copy
from tempfile import mktemp
from time import strftime

import logger

#################################################################
# auxillary functions
#################################################################

global DEBUG

def debug(message, threshold = 1):
	"""
	Routine for printing debugging info messages.
	"""
	 
	if threshold > DEBUG:
		return
	
	print "%s %s\n" %(strftime("%d%m%y %H:%M:%S"), message)


################################################################
# CConfigFile
class CConfigFile:
	"""
	CConfigFile is class for storing and loading configuration settings from file.
	
	Attributes:
		file          -- file with configuration, specified by a file handler or
		                 path it will raise an exception IOError if opening file
						 failed
		config_store  -- optional argument with CConfigStore object
		config_parser -- instance of ConfigParser
	
	Methods:
		load  -- loads configuration from file
		store -- stores configuration to file (do not directly use)
		flush -- actualize content of the configuration file with actual
		         settings (use this)
	"""
	
	def __init__(self, file, aConfig_store):
		if type(file) is types.StringType:
			self.file = open(file, "rw")
		elif type(file) is types.FileType:
			self.file = file
		else:
			raise IOError
		
		self.config_store = aConfig_store
		
		self.config_parser = ConfigParser.ConfigParser()
	
	def load(self):
		"""
		Loads configuration from config file
		"""
		
		self.config_parser.readfp(self.file)
				
		self.__readConfig()
		
	def store(self, aConfig_store = None):
		"""
		Stores configuration under description.
		"""
		
		self.config_store = aConfig_store
		
		self.flush()
	
	def flush(self):
		"""
		Writes changes to the file.
		"""
		
		# create temprorary file
		tmp_filename = mktemp()
		filename = self.file.name
		
		self.__writeConfig()
		
		self.config_parser.write(open(tmp_filename, "w+"))
		
		copy(tmp_filename, filename)
	
	def __string2BuildInType(self, str_value = ""):
		if str_value == "None":
			return None
		
		elif str_value == "True":
			return True
		elif str_value == "False":
			return False

		elif str_value == "*":
			return []

		elif sre.match("^\[.*\]", str_value):
			str_value = sre.sub("[\[\]]", "", str_value)
			val_list  = sre.split(",", str_value)
			
			ret_list = []
			if val_list == ['']:
				return ret_list
			else:
				for v in val_list:
					ret_list.append(self.__string2BuildInType(v))
			
			return ret_list
					
		elif type:
			try:
				return(int(str_value))
			except ValueError:
				pass
				
		if len(str_value) > 1:
			str_value = str_value.lstrip()
			if  str_value[0] == '\'' and str_value[len(str_value) - 1] == '\'':
				str_value = str_value[1:-1]
			elif str_value[0] == '"' and str_value[len(str_value) - 1] == '"':
				str_value = str_value[1:-1]
			
		return str_value
	
	def __string2CrontabType(self, str_value = ""):
		if str_value == "None" or str_value == "":
			return None
			
		elif sre.search(",", str_value):
			val_list = sre.split(",", str_value)
			
			ret_list = []
			for v in val_list:
				ret_list.append(self.__string2CrontabType(v))
			
			return CCrontabListValue(ret_list)
			
		elif sre.search("-", str_value):
			val_list = sre.split("-", str_value)
			if len(val_list) != 2:
				raise ValueError, str_value
				
			v1 = self.__string2BuildInType(val_list[0])
			v2 = self.__string2BuildInType(val_list[1])
			
			return CCrontabIntervalValue(v1, v2)
			
		elif sre.search("/", str_value):
			val_list = sre.split("/", str_value)
			if len(val_list) != 2:
				raise ValueError, str_value
			
			v1 = self.__string2BuildInType(val_list[0])
			v2 = self.__string2BuildInType(val_list[1])

			return CCrontabStepValue(v1, v2)
		
		return CCrontabSingleValue(self.__string2BuildInType(str_value))

	# This function is needed to convert lists to strings properly,
	# because str() on list containing string can corrupt encoding of the string.
	def __builtin_type_to_string(self, value = ""):

		if type(value) == type("string"):
			return "'" + value + "'"
		elif type(value) == type([]):
			s = ""
			for val in value :
				s += self.__builtin_type_to_string(val)
				s += ", "				

			return "[" +  s[:-2] + "]"
		else:
			return str(value)
			
	def __writeConfig(self):
		# clear config_parser
		for s in self.config_parser.sections():
			self.config_parser.remove_section(s)
			
		for (n, v)  in inspect.getmembers(self.config_store):
			if n[:2] != "__":
				vals = getattr(v, 'values')

				self.config_parser.add_section(n)
				
				for k in vals.keys():
					(item_value, item_cb) = vals[k]
					self.config_parser.set(n, k, self.__builtin_type_to_string(item_value))

	def __readConfig(self):
		config_store_members = dict(inspect.getmembers(self.config_store))

		for sect in self.config_parser.sections():
			options = self.config_parser.options(sect);
			items   = []
			for opts in options:
				val = self.config_parser.get(sect, opts);
				tupple = (opts, val);
				items.append(tupple);
			vals = dict(items)
			
			# convert values
			for k, v in vals.iteritems():
				v = self.__string2BuildInType(v)
				# eat also old config format for scheduler items
				if (( sect == 'test_sched_config' or sect == 'update_sched_config' ) and
				    ( k == 'hours' or k == 'mins' ) and
				    ( type(v) == int)):
					v = [v]
				config_store_members[sect].values[k] = (v, [])	

################################################################
# CAvgGuiConfig
class CAvgGuiConfig:
	"""
	CAvgGuiConfig is base class for holding AVG GUI settings.
	
	Attributes:
		values -- dictionary { name : tuple },
		          o name is string
				  o tuple consists of (value, callback)
	
	Methods:
		set_value -- set setting with value
		get_value -- return value of specified setting
		connect   -- callback with specific value, when this value changed, this
		             callback will be called with appropriate value
	"""
	
	def __init__(self):
		self.values   = {}
		self.__logger = logger.CLogger()

	def set_value(self, name = None, value = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		self.values[name] = (value, cbs)
		
		# call callbacks with new value
		for cb in cbs:
			cb(value)
		
		self.__logger.debug("Setting value '%s' to name '%s'." % (value, name))

	def get_value(self, name = None):
		try:
			(v, c) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		return v
		
	def connect(self, name = None, cb = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		cbs.append(cb)
		self.values[name] = (v, cbs)
	
	def disconnect(self, name = None, cb = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
		
		cbs.remove(cb)
		self.values[name] = (v, cbs)
		
################################################################
# CTestConfig
class CTestConfig(CAvgGuiConfig):
	"""
	CTestConfig is class for holding AVG GUI test settings.
	
	Methods:
		__str__ -- operator str() return cmd line string from values
	"""

	def __init__(self):
		CAvgGuiConfig.__init__(self)
		
		self.values = {
			'heur'      : (False, []),
			'rt'        : (False, []),
			'smart'     : (False, []),
			'ext'       : ([], []),
			'noext'     : ([], []),
			'arc'       : (False, []),
			'arcw'      : (False, []),
			'rtw'       : (False, []),
			'macrow'    : (False, []),
			'pwdw'      : (False, []),
			'changew'   : (False, []),
			'ignlocked' : (False, []),
			'repok'     : (False, []),
			'_use_def'  : (False, []),
			'def_files' : ([], [])
		}
		
	def __str__(self):
		tmp_opts = []
		for k in self.values.keys():
			(v, c) = self.values[k]
			if v == True and k[0] != "_":
				tmp_opts.append("-" + k)

		(v, c) = self.values['ext']
		if v:
			tmp_opts.append("-ext=%s" % ":".join(v))
		
		(v, c) = self.values['noext']
		if v:
			tmp_opts.append("-noext=%s" % ":".join(v))

		return " ".join(tmp_opts)

		
################################################################
# CUpdateConfig
class CUpdateConfig(CAvgGuiConfig):
	"""
	CUpdateConfig is class for holding AVG GUI update settings.
	
	Methods:
		__str__ -- operator str() return cmd line string from values
	"""
	def __init__(self):
		CAvgGuiConfig.__init__(self)

		self.values = {
			'no-diff'        : (False, []),
			'no-backup'      : (False, []),
			'no-log'         : (False, []),
			'no-daemons'     : (False, []),
			'complete'       : (False, []),
			'priority'       : (2, []),
			'server_name'    : ([], []),
			'server_url'     : ([], []),
			'_use_proxy'     : (False, []),
			'proxy_url'      : ("", []),
			'proxy_port'     : (0, []),
			'use_proxy_auth' : (False, []),
			'proxy_user'     : ("", []),
			'proxy_pass'     : ("", []),
			'bin_dir'        : ("", []),
			'upd_src'        : ("o", []), # source of update: o|p|f ... online(option -o) | PATH(option -f) | LIST(option -f)
			'upd_path'       : ("", []),  # PATH to update source files (option -f)
			'upd_files'      : ("", []),  # string representing LIST of update source files (option -f)
			}

	def __str__(self):
		tmp_opts = []
		for k in self.values.keys():
			(v, c) = self.values[k]
			if v == True and k[0] != "_":
				tmp_opts.append("--"+k)

		(v, c) = self.values['priority']
		tmp_opts.append("--priority %s" % v)

		(v, c) = self.values['upd_src']
		if v == 'o':
			tmp_opts.append("--online")
		# WARNING: must be the last in tmp_opts because the PATH|FILES must be last in cmdline
		elif v == 'p':
			tmp_opts.append("--offline %s" % self.values['upd_path'][0])
		elif v == 'f':
			tmp_opts.append("--offline %s" % " ".join(self.values['upd_files'][0]))	
		return " ".join(tmp_opts)
		
		


################################################################
# CTestResultsConfig
class CTestResultsConfig(CAvgGuiConfig):
	"""
	CTestResultsConfig is class for holding AVG GUI test results settings.
	
	"""

	def __init__(self):
		CAvgGuiConfig.__init__(self)

		self.values = {
			'num_saved' : (20, []),
			'del_older' : (30, []),
			'res_path'  : ("", [])
		}


################################################################
# CLicenseConfig
class CLicenseConfig(CAvgGuiConfig):
	"""
	CLicenseConfig is class for holding AVG GUI license settings.
	"""
	
	def __init__(self):
		CAvgGuiConfig.__init__(self)

		self.values = {
			'lic_num'   : ("", []),
			'user_name' : ("", []),
			'company'   : ("", [])
		}


################################################################
# CSchedulerConfig
class CSchedulerConfig(CAvgGuiConfig):
	"""
	CSchedulerConfig is base class for holding AVG GUI scheduler settings.
	"""
	
	def __init__(self):
		CAvgGuiConfig.__init__(self)

		# empty list stands for all values (like '*' in crontab file)
		self.values = {
			'mins'     : ([0], []),
			'hours'    : ([8], []),
			'days'     : ([], []),
			'months'   : ([], []),
			'd_of_mon' : ([], []),
			'run'      : (False, []),
			'at_reboot': (False, []),
			'max_delay': (0, [])
		}


	def get_cron_time_str(self):
		fields = ['mins', 'hours', 'd_of_mon', 'months', 'days']
		field_values = []
		for field in fields:
			if self.values[field][0] == []:
				field_values.append('*')
			else:
				field_values.append(','.join([str(day) for day in self.values[field][0]]))
		return ' '.join(field_values)		

	# TODO: UNUSED	
	def __str__(self):
		tmp_values = {}
		for k in self.values.keys():
			(v, c) = self.values[k]
			tmp_values[k] = v
			
		return ("%s %s %s %s %s" % (tmp_values['mins'], tmp_values['hours'],
									tmp_values['days'], tmp_values['months'],
									tmp_values['d_of_mon']))

		
################################################################
# CTestSchedulerConfig
class CTestSchedulerConfig(CSchedulerConfig):
	"""
	CTestSchedulerConfig is class for holding AVG GUI test scheduler settings.
	"""

	def __init__(self):
		CSchedulerConfig.__init__(self)

		self.values['def_files'] = ([], [])
		

################################################################
# CUpdateSchedulerConfig
class CUpdateSchedulerConfig(CSchedulerConfig):
	"""
	CUpdateSchedulerConfig is class for holding AVG GUI update scheduler settings.
	"""
	def __init__(self):
		CSchedulerConfig.__init__(self)


################################################################
# CConfigStorage
class CConfigStorage:
	"""
	CConfigStorage is class for encapsulating different Config classes
	"""
	
	test_config         = CTestConfig()
	update_config       = CUpdateConfig()
	license_config      = CLicenseConfig()
	test_res_config     = CTestResultsConfig()
	test_sched_config   = CTestSchedulerConfig()
	update_sched_config = CUpdateSchedulerConfig()


################################################################
# Classes for representing crontab values
#
# CCrontabSingleValue, CCrontabIntervalValue, CCrontabListValue,
# CCrontabStepValue
# NOTE: UNUSED
class CCrontabSingleValue:
	"""
	CCrontabValue is base class for holding single value.
	"""
	
	def __init__(self, value = None):
		self.data = value
		
	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return str(self.data)

class CCrontabIntervalValue:
	"""
	CCrontabIntervalValue is class for holding interval of values.
	"""
	
	def __init__(self, mini = None, maxi = None):
		self.data = (mini, maxi)
	
	def __str__(self):
		return self.__repr__()
	
	def __repr__(self):
		return '%s-%s' % self.data

class CCrontabListValue:
	"""
	CCrontabListValue is class for holding list of value.
	"""
	
	def __init__(self, value_list = None):
		self.data = value_list
	
	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return str.join(self.data, ",")
		
class CCrontabStepValue:
	"""
	CCrontabListValue is class for holding steps of value.
	"""
	
	def __init__(self, value = None, step = None):
		self.value = value
		self.step  = step
	
	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return '%s/%s' % (self.value, self.step)

################################################################
# testing code
if __name__ == "__main__":

	# turn on debug messages
	DEBUG = 0;
	
	# create new instance of gui configu
	avg_conf = CConfigStorage
	
	# create new config file
	try:
	
		conf_file = CConfigFile("default1.cfg", avg_conf)
		conf_file.load()
		
		
		avg_conf.test_config.set_value("heur", True)

		cr_int = CCrontabIntervalValue(0, 10)

		avg_conf.test_sched_config.connect("mins", debug)
		avg_conf.test_sched_config.connect("mins", debug)
		avg_conf.test_sched_config.set_value("mins", cr_int)
		avg_conf.test_sched_config.disconnect("mins", debug)
		avg_conf.test_sched_config.set_value("mins", cr_int)

		#conf_file.store(avg_conf)

	except IOError, what:
		print "** ERR: %s" % what

		
