"""
Module for managing scheduler features of GUI

Copyright (C) GRISOFT, s.r.o.
"""

import config
import logger

import os
import sre

from tempfile import mktemp

################################################################
# CCrontabEdit
class CCrontabEdit:
	"""
	CCrontabEdit is class for editing crontab of the user.
	
	Attributes:
		conf   -- CConfigStorage
		
	Methods:
		update -- actualize content of crontab with values from config
	"""
	
	def __init__(self):
		self.conf     = config.CConfigStorage
		self.__logger = logger.CLogger()
	
	def __readCrontab(self):
		crontab_cmd_output = os.popen("crontab -l 2>/dev/null", "r", 1)
		
		# jump over comments
		line = crontab_cmd_output.readline()
		while sre.match("^#\s.*", line):
			line = crontab_cmd_output.readline()
		
		return [line] + crontab_cmd_output.readlines()
	
	def __scan_cron_cmd(self):
		try:
			res_path = self.conf.test_res_config.get_value("res_path")
			
			res_filename = res_path + '/testreport.`date  +\%s`'
			
			params   = str(self.conf.test_config)
			paths    = self.conf.test_sched_config.get_value("def_files")
			paths    = " ".join(paths)
		except AttributeError, what:
			self.__logger.debug("AttributeError: %s" % what)
			pass
			
		avgscan_cmd = "/usr/bin/avgscan -report $FILENAME"

		cmdline = [avgscan_cmd, params, paths]
		cmdline = " ".join(cmdline)
		
		cron_cmd  = "(FILENAME=" + res_filename + "; " + cmdline + "|| mv $FILENAME $FILENAME.)"

		return cron_cmd

	def __scan_rule_rb(self):
		# at reboot rule
		return "@reboot" + " " + self.__scan_cron_cmd() + "\n"
	
	def __scan_rule(self):
		cron_time = self.conf.test_sched_config.get_cron_time_str()
		return cron_time + " " + self.__scan_cron_cmd() + "\n"

	def __scan_rule_d(self):
		# delayed rule
		cron_time = self.conf.test_sched_config.get_cron_time_str()
		max_delay = self.conf.test_sched_config.get_value("max_delay")
		return cron_time + " at now + $(($RANDOM `echo \%` " + str(max_delay) + " + 1)) minutes > /dev/null 2>&1 %" + self.__scan_cron_cmd()  + "\n"
	
	def __updt_cron_cmd(self):
		params = str(self.conf.update_config)
		updt_cmd = "/usr/bin/avgupdate -n "		
		cron_cmd  = updt_cmd + params
		return cron_cmd

	def __updt_rule_rb(self):
		# at reboot rule
		return "@reboot" + " " + self.__updt_cron_cmd() + "\n"

	def __updt_rule(self):
		cron_time = self.conf.update_sched_config.get_cron_time_str()
		return cron_time + " " + self.__updt_cron_cmd() + "\n"

	def __updt_rule_d(self):
		# delayed rule
		cron_time = self.conf.update_sched_config.get_cron_time_str()
		max_delay = self.conf.update_sched_config.get_value("max_delay")
		return cron_time + " at now + $(($RANDOM `echo \%` " + str(max_delay) + " + 1)) minutes > /dev/null 2>&1 %" + self.__updt_cron_cmd()  + "\n"
		
	def update(self):
		"""
		Actualize content of crontab with values from config.
		"""

		# prepare material
		crontab = self.__readCrontab()		
		self.__logger.debug("old crontab: %s" % crontab)

		new_cron_rules = []

		ch_scan_per = False
		ch_scan_reb = False
		ch_upd_per = False
		ch_upd_reb = False		

		# periodical test
		if self.conf.test_sched_config.get_value("run"):
			if self.__test_files_selected():
				new_cron_rules.append("#AVGSCAN-GUI\n")
				if self.conf.test_sched_config.get_value("max_delay") > 0:
					# randomly delayed scan
					new_cron_rules.append(self.__scan_rule_d())
				else:	
					new_cron_rules.append(self.__scan_rule())
				ch_scan_per = True
			else:
				self.__logger.warn("scheduled (periodical) test settings were not changed, reason: 0 paths to scan selected")			     
		else:
			ch_scan_per = True

		# test at reboot
		if self.conf.test_sched_config.get_value("at_reboot"):
			if self.__test_files_selected():
				new_cron_rules.append("#AVGSCAN-GUI\n")
				new_cron_rules.append(self.__scan_rule_rb())				
				ch_scan_reb = True
			else:
				self.__logger.warn("scheduled (at reboot) test settings were not changed, reason: 0 paths to scan selected")			     
		else:
			ch_scan_reb = True

		# periodical update
		if self.conf.update_sched_config.get_value("run"):
			if self.__upd_src_selected():
				new_cron_rules.append("#AVGUPDATE-GUI\n")
				if self.conf.update_sched_config.get_value("max_delay") > 0:
					# randomly delayed update
					new_cron_rules.append(self.__updt_rule_d())
				else:
					new_cron_rules.append(self.__updt_rule())				
				ch_upd_per = True
			else:
				self.__logger.warn("scheduled (periodical) update settings were not changed, reason: source files or path not selected")
		else:
			ch_upd_per = True

		# update at reboot
		if self.conf.update_sched_config.get_value("at_reboot"):
			if self.__upd_src_selected():
				new_cron_rules.append("#AVGUPDATE-GUI\n")
				new_cron_rules.append(self.__updt_rule_rb())				
				ch_upd_reb = True
			else:
				self.__logger.warn("scheduled (at reboot) update settings were not changed, reason: source files or path not selected")
		else:
			ch_upd_reb = True			
			
		self.__logger.debug("new cron rules: %s" % new_cron_rules)

		# edit crontab
		new_crontab = []
		scan_line = ''
		upd_line = ''
		for line in crontab:
			# second line of avggui cron record
			if scan_line:
				if sre.match("^@reboot", line):
					if not ch_scan_reb:
						new_crontab.append(scan_line)
						new_crontab.append(line)

				elif not ch_scan_per:
					new_crontab.append(scan_line)
					new_crontab.append(line)

				scan_line = ''	
				continue
			# second line of avggui cron record
			if upd_line:
				if sre.match("^@reboot", line):
					if not ch_upd_reb:
						new_crontab.append(upd_line)
						new_crontab.append(line)

				elif not ch_upd_per:
					new_crontab.append(upd_line)
					new_crontab.append(line)

				upd_line = ''	
				continue

			# first line of avggui cron record
			if sre.match("^#AVGSCAN-GUI", line):
				scan_line = line
				continue
			# first line of avggui cron record
			elif sre.match("^#AVGUPDATE-GUI", line):
				upd_line = line
				continue
			else:
				new_crontab.append(line)
				
		# temporary file
		tmp_filename = mktemp()
		tmp_file = open(tmp_filename, "w+")
		
		tmp_file.writelines(new_crontab + new_cron_rules)
		tmp_file.flush()
		
		os.system("crontab " + tmp_filename)

		
	def __upd_src_selected(self):
		return (
			(self.conf.update_config.get_value('upd_src') == 'o')
			or
			(self.conf.update_config.get_value('upd_src') == 'p'
			 and self.conf.update_config.get_value('upd_path') != '')
			or
			(self.conf.update_config.get_value('upd_src') == 'f'
			 and self.conf.update_config.get_value('upd_files') != [])
			)
	def __test_files_selected(self):	
		return self.conf.test_sched_config.get_value('def_files') != []

################################################################
# testing code
if __name__ == "__main__":

	try:
		conf = config.CConfigStorage
		conf.test_config.set_value("repok", True)

		cron_edit = CCrontabEdit()
		cron_edit.update()
		
	except IOError, what:
		pass
	except AttributeError, which:
		pass
		
