r"""Classes for file system hierarchy listing

This module contains modified gtk classes TreeStore and TreeView for
file system hierarchy listing.

Copyright (C) GRISOFT, Inc.
"""

try: 
	from sys import exit
	import pygtk
	pygtk.require('2.0')
	import gtk
	from gtk import TreeStore, TreeView, TreeViewColumn, CellRendererText, CellRendererPixbuf, CellRendererToggle,   Window, WINDOW_TOPLEVEL, main
	from gtk.gdk import Pixbuf, pixbuf_new_from_xpm_data
	from stat import *
	import os
	import sre
	import messagebox
	import misc
	
	import logger
except Exception, detail:
	print detail
	exit(1)

################################################################
# CEntry
class CEntry:
	"""Contains information about one directory entry.
	   Information are stored in the 'info' dictionary.
	   
	   pathname  - full entry pathname
	   type      - contains one of the following:
	               reg: regular file
				   dir: directory
				   lnk: symlink
				   blk: block device
				   chr: character device
				   fifo: fifo
				   sock: socket"""

	def __init__(self, pathname, type):
		self.info = {
			'pathname': pathname,
			'type': type
		}

################################################################
# CDirEntries
class CDirEntries:
	"""Provides enhanced functionality of built-in listpath(...) function.
	   Main method get_entries returns a list containing the names of entries
	   in the directory. Following properties could be set up to customize
	   the GetEntries function behaviour.

	   hidden    - include hidden files
	   empty     - include empty directories
	   access    - include only entries to which user has access
	   directory - include directories
	   file      - include regular files
	   charDev   - include charakter devices
	   blockDev  - include block devices
	   fifo      - include fifo files
	   socket    - include unix socket files
	   sort      - sort names

	   Properties could be set up via 'pref' dictionary.
	   Example: obj.pref['Hidden'] = False
	            Where obj is instance of DirEntries class."""
				
	def __init__(self):
		"""Sets up default behaviour."""

		self.pref = {
			'hidden': True,
			'empty': True,
			'access': True,
			'directory': True,
			'file': True,
			'charDev': False,
			'blockDev': False,
			'fifo': False,
			'symlink': False,
			'socket': False,
			'sort': True
		}
		
		self.__logger = logger.CLogger()

	def __empty(self, dir):
		"""Returns True if the directory specified with argument 'dir' is empty
		   or is not accessible. Otherwise returns False"""
		   
		items = []
		try:
			items = os.listdir(dir)
		except OSError, detail:
			self.__logger.warn(detail)
			return True
		
		if items == []:
			return True
		else:
			return False
		
	
	def __accessible(self, path, mode):
		"""Returns True if the path is accessible. Otherwise returns False"""
		
		if S_ISDIR(mode):
			return os.access(path, os.X_OK | os.R_OK)
		else:
			return os.access(path, os.R_OK)
		
	def __hidden(self, path):
		"""Returns True if the path is hidden. Otherwise returns False"""
		
		name = os.path.basename(path)
		if name[0] == '.':
			return True
		else:
			return False
		
	def __entry_cmp(self, entry1, entry2):
		"""Internal compare method for CEtry objects"""
		
		return cmp(entry1.info['pathname'], entry2.info['pathname'])

	def get_entries(self, dir):
		"""Returns a list containing information about entries in the directory."""

		items, dirs, files  = [], [], []
		try:
			items = os.listdir(dir)
		except OSError, detail:
			self.__logger.warn(detail)
			return []

		for item in items:
			path = os.path.join(dir,item)
			try:
				stat_res = os.stat(path)
			except OSError, detail:
				self.__logger.warn(detail)
				continue
			mode = stat_res.st_mode

			if self.pref['access'] and not self.__accessible(path, mode):				# check if directory entry should be added
				continue

			if not self.pref['hidden'] and self.__hidden(path):
				continue

			if self.pref['file'] and S_ISREG(mode):
				files.append(CEntry(path,'reg'))
			elif self.pref['directory'] and S_ISDIR(mode):
				if self.pref['empty'] or not self.__empty(path):
					dirs.append(CEntry(path,'dir'))
			elif self.pref['symlink'] and S_ISLNK(mode):
				files.append(CEntry(path,'lnk'))
			elif self.pref['blockDev'] and S_ISBLK(mode):
				files.append(CEntry(path,'bkl'))
			elif self.pref['charDev'] and S_ISCHR(mode):
				files.append(CEntry(path,'chr'))
			elif self.pref['socket'] and S_ISSOCK(mode):
				files.append(CEntry(path,'sock'))
			elif self.pref['fifo'] and S_ISFIFO(mode):
				files.append(CEntry(path,'fifo'))

		if self.pref['sort']:
			dirs.sort(self.__entry_cmp)
			files.sort(self.__entry_cmp)
		
		return dirs + files

################################################################
# CRow
class CRow:
	"""Holds details about one row in CStore."""
	
	def __init__(self, pathname, type, cached):
		self.info = {
			'pathname': pathname, # full pathname
			'type'    : type,     # reg, dir, lnk, chr, blk, fifo, sock
			'cached'  : cached    # flag that content of directory in already readed
		}

	
	
################################################################
# CStore
class CStore(TreeStore):
	"""Class holds file system hierarchy, that is created dynamically."""

	def __init__(self, pathnames, selected = False, expand = False, w_message = False):
		TreeStore.__init__(self, int, Pixbuf, str, str, object)
		self.dir_entries = CDirEntries()
		
		self.__logger    = logger.CLogger()

		self.column = {
			'checkbox'   : 0, # value of checkbox 0 => unchecked, 1 = > checked
			'icon'       : 1, # icon displayed with pathname
			'name'       : 2, # basename of pathname
			'description': 3, # text that may describe pathname
			'info'       : 4  # CRow info
		}

		self.xpm = {
			'dir_collapsed':
				["16 16 16 1",
			     "       c None",
			     ".      c #f4f7e4",
			     "X      c #dee4b5",
			     "o      c #e1e7b9",
			     "O      c #c6cba4",
			     "+      c #dce2b8",
			     "@      c #e9e9ec",
			     "#      c #d3d8ae",
			     "$      c #d8daca",
			     "%      c #b2b2b5",
			     "&      c #767862",
			     "*      c #e3e6c3",
			     "=      c #1b1b1a",
			     "-      c #939684",
			     ";      c #555555",
			     ":      c #000000",
			     "                ",
			     "                ",
			     "  ::::          ",
			     " :.@@O:         ",
			     ":-&&&&&:::::    ",
			     ":.@@@@@*$O#O=   ",
			     ":@*+XXXX+##O:   ",
			     ":.*#oooXXXXX:   ",
			     ":@+XoXXXXXX#:   ",
			     ":@*ooXXXXXX#:   ",
			     ":@**XXXXXXX#:   ",
			     ":@*XXXXXXXX%:   ",
			     ":$.*OOOOOO%-:   ",
			     " ;:::::::::::   ",
			     "                ",
			     "                "],
			'dir_expanded': 
			 	["16 16 16 1",
			     "       c None",
			     ".      c #a9ad93",
			     "X      c #60634d",
			     "o      c #dee4b5",
			     "O      c #9ca085",
			     "+      c #0c0d04",
			     "@      c #2f2f31",
			     "#      c #3b3d2c",
			     "$      c #c8cda2",
			     "%      c #e6e6e9",
			     "&      c #b3b5a5",
			     "*      c #80826d",
			     "=      c #292a1c",
			     "-      c #fefef6",
			     ";      c #8f937b",
			     ":      c #000000",
			     "                ",
			     "                ",
			     "  ::::          ",
			     " :-%%&:         ",
			     ":-;;;OX:::::    ",
			     ":-;;;;O;O;&.:   ",
			     ":-*X##@@@@@=#:  ",
			     ":%*+-%%ooooooO: ",
			     ":%X;%ooooooo.*: ",
			     ":.+-%oooooooO:  ",
			     ":*O-oooooooo*:  ",
			     ":O-oooooooo.:   ",
			     ":*-%$$$$$$OX:   ",
			     " :::::::::::    ",
			     "                ",
			     "                "],

			'file_reg': 
				["16 16 42 1",
				 "       c None",
				 ".      c #000000",
				 "+      c #FEFEFE",
				 "@      c #FDFDFD",
				 "#      c #E0E0E0",
				 "$      c #C1C1C1",
				 "%      c #F1F1F1",
				 "&      c #C3C3C3",
				 "*      c #FBFBFB",
				 "=      c #A8A8A8",
				 "-      c #ADADAD",
				 ";      c #767676",
				 ">      c #5D5D5D",
				 ",      c #404040",
				 "'      c #F0F0F0",
				 ")      c #E2E2E2",
				 "!      c #858585",
				 "~      c #4B4B49",
				 "{      c #161616",
				 "]      c #EFEFEF",
				 "^      c #EEEEEE",
				 "/      c #C5C5C5",
				 "(      c #ECECEC",
				 "_      c #C4C4C4",
				 ":      c #EBEBEB",
				 "<      c #EDEDED",
				 "[      c #EAEAEA",
				 "}      c #E8E8E8",
				 "|      c #C0C0C0",
				 "1      c #E9E9E9",
				 "2      c #E7E7E7",
				 "3      c #BFBFBF",
				 "4      c #E6E6E6",
				 "5      c #E5E5E5",
				 "6      c #BEBEBE",
				 "7      c #E4E4E4",
				 "8      c #BDBDBD",
				 "9      c #E3E3E3",
				 "0      c #BBBBBB",
				 "a      c #C2C2C2",
				 "b      c #BCBCBC",
				 "c      c #A0A0A0",
				 "   .........    ",
				 "  .++++++@#$.   ",
				 "  .+%%%%%%&*=.  ",
				 "  .+%%%%%%-;>,. ",
				 "  .+%%%%%')!~{. ",
				 "  .+%]]]]]]^^/. ",
				 "  .+]]]^^^^^(_. ",
				 "  .@^^^^^^^^:&. ",
				 "  .@<(((::[[[$. ",
				 "  .@:}}}}}}}}|. ",
				 "  .@[[11}}}223. ",
				 "  .@}444444456. ",
				 "  .@2244455778. ",
				 "  .@4555779990. ",
				 "  .a888bb0000c. ",
				 "   ...........  "]
		}

		self.pixmap = {
			'dir_collapsed': pixbuf_new_from_xpm_data(self.xpm['dir_collapsed']),
			'dir_expanded': pixbuf_new_from_xpm_data(self.xpm['dir_expanded']),
			'file_reg': pixbuf_new_from_xpm_data(self.xpm['file_reg'])
		}

		for pathname in pathnames:                         											# init CStore with given pathnames
			self.add_pathname(pathname, selected, expand, w_message)
			
	def get_iter_from_name(self, parent, name):
		"""Search for row with basename 'name' in all children of parent in same level.
		   Returns iter if row is found, otherwise None."""
		
		if parent == None:
			return None
			
		child = self.iter_children(parent)
		
		if child == None:
			return None

		row_info = self.get_value(child, self.column['info'])

		if row_info == None:																		# check fake row
			return None

		iter = None
		done = False
		
		while not done:																				# search all children of given parent
			if child == None:
				done = True
			else:
				if name == self.get_value(child, self.column['name']):
					iter = child
					done = True
				else:
					child = self.iter_next(child)

		return iter
	
	def __print_row_info(self, iter, user_data = None):
		"""Print row info."""
		
		check_box   = self.get_value(iter, self.column['checkbox'])
		icon        = self.get_value(iter, self.column['icon'])
		name        = self.get_value(iter, self.column['name'])
		description = self.get_value(iter, self.column['description'])
		info        = self.get_value(iter, self.column['info'])
		
		print ("%-3d %-20s %-30s %-20s %-60s %-10s %-3d") % (check_box, icon, name, description, \
		info.info['pathname'], info.info['type'], info.info['cached'])
		return
	
	def __check_row(self, iter, value):
		"""Set new value to checkbox."""
		
		self.set_value(iter, self.column['checkbox'], value)

		return

	def expand(self, parent, no_selected_children = False):
		"""Expand pathname defined by parent."""
		
		if parent == None:																			
			return

		row_info = self.get_value(parent, self.column['info'])

		if row_info.info['cached'] == True:															# this parent is already expaded
			return 

		if row_info.info['type'] != 'dir':															# only directories could be expanded
			return
			
		pathname = row_info.info['pathname']
		self.set_value(parent, self.column['icon'], self.pixmap['dir_expanded'])
		entries = self.dir_entries.get_entries(pathname) 
		row_info.info['cached'] = True
		for entry in entries:
			checkbox = 0
			if not no_selected_children:
				checkbox    = self.get_value(parent, self.column['checkbox'])						# init values for new row

			icon        = self.pixmap['file_reg']
			
			if entry.info['type'] == 'dir':
				icon = self.pixmap['dir_collapsed']

			name        = os.path.basename(entry.info['pathname'])
			description = None
			row_info    = CRow(entry.info['pathname'], entry.info['type'], False)
			
			child = self.append(parent, [checkbox, icon, name, description, row_info])

			if entry.info['type'] == 'dir':															# if new child entry is directory append to it new empty(fake) row
				self.append(child, [0, None, None, None, None])										# we want to see expander

		self.remove(self.iter_children(parent))														# remove fake row from parent

		return 
	
	def get_iter_from_pathname(self, pathname):
		"""Search for row in whole storage tree which contains pathname.
		   Returns 2-tuple (iter, list) where 'iter' is iter to the found row and
		   'list' is list of path items that are not saved in tree.
		   example: this pathname is saved in tree '/home/user' and we are searching
		            for '/home/user/mp3/sound'. Method will return (iter, ["mp3", "sound"])
					where 'iter' is iter for row that contains '/home/user' pathname"""
		
		if pathname == [] or pathname == None:
			return (self.get_iter_root(), None)

		pathname_splited = sre.split("/", pathname[1:])
		parent           = self.get_iter_root()
		index            = 0
		done             = False
		
		while not done:
			child = self.get_iter_from_name(parent, pathname_splited[index])

			if child == None:																		# entry name in parent was not found
				done = True
				child = parent
			elif index == len(pathname_splited) - 1:												# pathname is already added
				done = True
				index = index + 1
			else:																					# entry name was found, lets search child
				parent = child
				index = index + 1

		return (child, pathname_splited[index:])



	def __get_selected_iter(self, iter = None):
		"""
		Returns list of selected pathnames in subtree defined by iter.
		Note: Only the top selected pathnames are returned. All
		pathnames selected below are ignored.
		"""

		if iter == None:
			return []

		next = iter
 		result = []

		while next != None:
			
			checkbox = self.get_value(next, self.column['checkbox'])
			if checkbox:
				# skip items in subtree
				info   = self.get_value(next, self.column['info'])
				result = result + [info.info['pathname']]
			else:
				child  = self.iter_children(next)
				result = result + self.__get_selected_iter(child)
	
			next = self.iter_next(next)

		return result



	def __get_af_selected_iter(self, view, iter = None, filter_types=['dir']):
		"""
		Returns list of selected pathnames in subtree defined by iter.

		arguments:
		view -- treeview of model - only expanded rows are selected

		Note:
		Unlike in the __get_selected_iter() method, paths of all rows
		which aren't filtered on the basis of their type
		(filter_types argument) are returned.

		This is used typically for selecting (getting) only file paths.
		"""

		if iter == None:
			return []

		next = iter
 		result = []

		while next != None:
			
			checkbox = self.get_value(next, self.column['checkbox'])
			if checkbox:
				# skip items in subtree
				info   = self.get_value(next, self.column['info'])
				if not info.info['type'] in filter_types: # ignore filtered types
					result = result + [info.info['pathname']]
				#if (info.info['type'] != 'dir') or (view.row_expanded(self.get_path(next))):
				if (view.row_expanded(self.get_path(next))):
					# we traverse expanded checked directories because they can contain
					# displayed checked files
					child  = self.iter_children(next)
					result = result + self.__get_af_selected_iter(view, child, filter_types)	
			else:
				child  = self.iter_children(next)
				result = result + self.__get_af_selected_iter(view, child, filter_types)

			next = self.iter_next(next)

		return result

	def add_pathname(self, pathname, selected = False, expand = False, w_message = False):
		"""Add pathname to store."""
		
		if self.get_iter_root() == None:															# store is empty, add root row
			checkbox    = 0		
			
			if pathname == "/":
				checkbox = selected

			icon        = self.pixmap['dir_collapsed']
			name        = "/"
			description = None
			row_info    = CRow("/", 'dir', False)
			
			child       = self.append(None,[checkbox, icon, name, description, row_info])
			
			if expand == True:																		# expand root
				self.append(child, [0, None, None, None, None])
				self.expand(self.get_iter_root())
				row_info = self.get_value(self.get_iter_root(), self.column['info'])
				row_info.info['cached'] = True
				self.set_value(self.get_iter_root(), self.column['icon'], self.pixmap['dir_expanded'])
			else:
				self.append(child, [0, None, None, None, None])										# add fake row to root

		if pathname == "/":																			# root is already added
			return

		pathname = os.path.normpath(pathname)
		iter, pathname_rest = self.get_iter_from_pathname(pathname)

		if len(pathname_rest) == 0:																	# path name is already added
			if selected:
				self.set_value(iter, self.column['checkbox'], 1)
			return 
		
		for name in pathname_rest:																	# add new pathname to store
			self.expand(iter)
			iter = self.get_iter_from_name(iter, name)
			if iter == None:
				if w_message:
					message_window = messagebox.CAvgMessageOkDialog(aParent = gtk.glade.XML("../glade/avggui.glade").get_widget("wSelectObjects"),
											aType           = gtk.MESSAGE_WARNING,
											aFlags          = gtk.DIALOG_MODAL,
											aMessage_format = "path '%s' can't be tested" % pathname,
											aLogMessage     = True)

					message_window.start()
				else:	
					#self.__logger.warn("path '%s' can't be tested" % pathname)
					pass

				break;

		if iter != None:
			self.set_value(iter, self.column['checkbox'], selected)									# set checkbox value to the last directory entry
			self.check_parents(iter)

		if expand == True:																			# expand last pathname item
			self.expand(iter)
		
		return
		
	def foreach_iter(self, parent, func, user_data = None):
		"""For each iter in subtree whose top is defined with iter 'parent'
		   call user funciton 'func' with user data 'user_data'."""
		   
		if parent == None:
			return
		
		func(parent, user_data)

		child = self.iter_children(parent)
		if parent == None:
			return

		func(parent, user_data)
		child = self.iter_children(parent)
		
		if child == None:
			return

		child_info = self.get_value(child, self.column['info'])

		if child_info == None:																		# check fake row
			return

		while child != None:
			self.foreach_iter(child, func, user_data)
			child = self.iter_next(child)

		return
		
	def dump(self):
		"""Print content of store."""
		
		self.foreach_iter(self.get_iter_root(), self.__print_row_info)
		return

	def check_all(self, iter, value):
		"""Set new checkbox value for iter and all children."""
		
		self.foreach_iter(iter, self.__check_row, value)
		return

	def check_parents(self, iter):
		"""Check if all rows in level specified with iter are checked.
		   If so, the parent is checked too. This is done with recursion to the top."""
		
		if iter == None:
			return
			
		parent = self.iter_parent(iter)
		
		if parent == None:
			return
			
		child  = self.iter_children(parent)
		go_on  = True
		done   = False

		while child != None and not done:
			checkbox = self.get_value(child, self.column['checkbox'])
			if not checkbox:
				done = True
				go_on = False
			else:
				child = self.iter_next(child)

		if go_on:															# all rows in this level are checked 
			self.set_value(parent, self.column['checkbox'], 1)
			self.check_parents(parent)
				
		return
	
	def uncheck_parents(self, iter):
		"""Uncheck parent row if not all children rows are checked. This is done
		   with recursion to the top."""
		   
		if iter == None:
			return

		parent   = self.iter_parent(iter)
		
		if parent == None:
			return
			
		checkbox = self.get_value(parent, self.column['checkbox'])

		if checkbox:
			self.set_value(parent, self.column['checkbox'], 0)
			self.uncheck_parents(parent)
			
		return
	
	def get_selected(self):
		"""Returns list of selected pathnames in store.
		   None: Only the top selected pathnames are returned. All
		         pathnames selected below are ignored."""
				 
		return self.__get_selected_iter(self.get_iter_root())
	
	def get_af_selected(self, view, filter_types):
		"""
		Returns list of selected pathnames in store.

		Note:
		Unlike in the __get_selected() method, paths of all rows
		which aren't filtered on the basis of their type
		(filter_types argument) are returned.

		This is used typically for selecting (getting) only file paths.
		"""

		return self.__get_af_selected_iter(view, self.get_iter_root(), filter_types)

	def init(self, pathnames, selected = False, expand = False, w_message = False):
		"""Init store with pathnames. All previous content of store is deleted."""

		self.clear()
		for pathname in pathnames:
			self.add_pathname(pathname, selected, expand, w_message)

################################################################
# CView
class CView(TreeView):
	"""
	This class is used to display CStore data.

	arguments:
	w_message - display warning messagebox in case one of pathnames is not found	
	"""
	def __init__(self, pathnames, selected = False, expand = False, w_message = False ):
		TreeView.__init__(self)
		self.model = CStore(pathnames, selected, expand, w_message)
		self.set_model(self.model)

		self.column_path        = TreeViewColumn("Path name")									# column for checkbox, icon and path basename
		#column_description = TreeViewColumn("Description")										# column for description
		
		self.renderer_checkbox    = CellRendererToggle()										# cellrenderers
		self.renderer_icon        = CellRendererPixbuf()
		self.renderer_name        = CellRendererText()
		#renderer_description = CellRendererText()

		self.column_path.pack_start(self.renderer_checkbox, False)											# add cellrenderers to columns
		self.column_path.pack_start(self.renderer_icon, False)
		self.column_path.pack_start(self.renderer_name, False)
		#column_description.pack_start(renderer_description, False)
		
		self.column_path.add_attribute(self.renderer_checkbox, "active", self.model.column['checkbox'])		# set columns
		self.column_path.add_attribute(self.renderer_icon, "pixbuf", self.model.column['icon'])
		self.column_path.add_attribute(self.renderer_name, "text", self.model.column['name'])
		#column_description.add_attribute(renderer_description, "text", self.model.column['description'])
		
		self.append_column(self.column_path)																# add columns to TreeView
		#self.append_column(column_description)

		self.expanded_handler = self.connect('row-expanded', self.__expand)									# register signals
		self.connect('row-collapsed', self.__collapsed)
		self.toggled_handler = self.renderer_checkbox.connect('toggled', self.__toggled)

		self.set_property('headers-visible', False)												# set properties
		self.set_property('rules-hint', True)
		self.renderer_checkbox.set_property('yalign', 0.5)
		self.renderer_checkbox.set_property('ypad', 0)
		self.renderer_icon.set_property('yalign', 0.5)
		self.renderer_icon.set_property('ypad', 0)
		self.renderer_name.set_property('yalign', 0.5)
		self.renderer_name.set_property('ypad', 0)

		self.column_path.set_cell_data_func(self.renderer_name, self.__filename_to_utf8)

		self.expand_pathnames(pathnames, expand)
		
	def __filename_to_utf8(self, treeviewcolumn, cell_renderer, tree_model, iter):
		filename = tree_model.get_value(iter, tree_model.column['name'])
		cell_renderer.set_property('text', misc.filename_to_utf8(filename)) 
		return

	def expand_pathnames(self, pathnames, expand = False):
		for pathname in pathnames:																# show selected pathnames
			self.__expand_to_pathname(pathname, expand)
		
	def __expand_to_pathname(self, pathname, expand = False):
		pathname = os.path.normpath(pathname)

		if pathname == "/":
			pathname_splited = []
			if not expand:
				return
		else:
			pathname_splited = sre.split("/", pathname[1:])
			if not expand:
				pathname_splited = pathname_splited[:-1]
		
		parent = self.model.get_iter_root()
		self.expand_row(self.model.get_path(parent), False)

		for name in pathname_splited:
			parent = iter = self.model.get_iter_from_name(parent, name)
			if iter == None:
				break;
			self.expand_row(self.model.get_path(iter), False)
			
	def __expand(self, treeview, iter, path, user_data = None):
		self.model.set_value(iter, self.model.column['icon'], self.model.pixmap['dir_expanded'])
		self.model.expand(iter)
		return

	def __collapsed(self, treeview, iter, path, user_data = None):
		self.model.set_value(iter, self.model.column['icon'], self.model.pixmap['dir_collapsed'])
		return
		
	def __toggled(self, cellrenderer, path, user_data = None):
		iter     = self.model.get_iter(path)
		checkbox = self.model.get_value(iter, self.model.column['checkbox'])

		if not checkbox:
			self.model.check_all(iter, 1)														# set new checkbox value to current iter and all children
			self.model.check_parents(iter)														# check if parent of iter should be changed
		else:
			self.model.check_all(iter, 0)
			self.model.uncheck_parents(iter)
		
		return

	def get_selected(self):
		"""Return selected pathnames. Wraps get_selected method from CStore"""

		return self.model.get_selected()

	def get_af_selected(self, filter_types):
		"""Return selected pathnames. Wraps get_af_selected method from CStore"""
		
		return self.model.get_af_selected(self, filter_types)	

	def init(self, pathnames, selected = False, expand = False, w_message = False):
		"""Sets new values. All previously saved pathnames are deleted."""
		
		self.model.init(pathnames, selected, expand, w_message)
		self.expand_pathnames(pathnames, expand)
		
class COneDirSelection(CView):
	def __init__(self, pathname, selected):
		CView.__init__(self, [], selected)
		self.checked_path = None
		
		self.model.dir_entries.pref['hidden']    = True
		self.model.dir_entries.pref['empty']     = True
		self.model.dir_entries.pref['access']    = True
		self.model.dir_entries.pref['directory'] = True
		self.model.dir_entries.pref['file']      = False
		self.model.dir_entries.pref['charDev']   = False
		self.model.dir_entries.pref['blockDev']  = False
		self.model.dir_entries.pref['fifo']      = False
		self.model.dir_entries.pref['symlink']   = False
		self.model.dir_entries.pref['socket']    = False
		self.model.dir_entries.pref['sort']      = True

		self.init([pathname], selected)

		if selected:
			iter, path_rest = self.model.get_iter_from_pathname(pathname)
			self.checked_path = self.model.get_path(iter)
		
		self.disconnect(self.expanded_handler)
		self.renderer_checkbox.disconnect(self.toggled_handler)
		
		self.renderer_checkbox.connect('toggled', self.__select)
		self.connect('row-expanded', self.__expand)

	def __select(self, cellrenderer, path, user_data = None):
		iter = self.model.get_iter(path)
		checkbox = self.model.get_value(iter, self.model.column['checkbox'])

		if not checkbox:
			if self.checked_path:
				old_iter = self.model.get_iter(self.checked_path)
				self.model.set_value(old_iter, self.model.column['checkbox'], 0)

			self.checked_path = path
			self.model.set_value(iter, self.model.column['checkbox'], 1)
		
		return

	def __expand(self, treeview, iter, path, user_data = None):
		self.model.set_value(iter, self.model.column['icon'], self.model.pixmap['dir_expanded'])
		checkbox = self.model.get_value(iter, self.model.column['checkbox'])
		self.model.expand(iter, True)
		return

class COnePathSelection(CView):
	"""
	Provides widget (based on TreeView) for selection fo one path.
	
	Default is selection of directory.
	""" 
	def __init__(self, pathname, selected,
		     set_props=['hidden', 'empty', 'access', 'directory', 'sort'],
		     unset_props=['file', 'charDev', 'blockDev', 'fifo', 'symlink', 'socket']):
		CView.__init__(self, [], selected)
		self.checked_path = None
		
		self.model.dir_entries.pref['hidden']    = True
		self.model.dir_entries.pref['empty']     = True
		self.model.dir_entries.pref['access']    = True
		self.model.dir_entries.pref['directory'] = True
		self.model.dir_entries.pref['file']      = False
		self.model.dir_entries.pref['charDev']   = False
		self.model.dir_entries.pref['blockDev']  = False
		self.model.dir_entries.pref['fifo']      = False
		self.model.dir_entries.pref['symlink']   = False
		self.model.dir_entries.pref['socket']    = False
		self.model.dir_entries.pref['sort']      = False

		for prop in set_props:
			self.set_prop(prop)
		for prop in unset_props:
			self.unset_prop(prop)
			
		self.init(pathname, selected)

		self.disconnect(self.expanded_handler)
		self.renderer_checkbox.disconnect(self.toggled_handler)
		
		self.renderer_checkbox.connect('toggled', self.__select)
		self.connect('row-expanded', self.__expand)

	def __select(self, cellrenderer, path, user_data = None):
		iter = self.model.get_iter(path)
		checkbox = self.model.get_value(iter, self.model.column['checkbox'])

		if not checkbox:
			if self.checked_path:
				old_iter = self.model.get_iter(self.checked_path)
				self.model.set_value(old_iter, self.model.column['checkbox'], 0)

			self.checked_path = path
			self.model.set_value(iter, self.model.column['checkbox'], 1)
		
		return

	def __expand(self, treeview, iter, path, user_data = None):
		self.model.set_value(iter, self.model.column['icon'], self.model.pixmap['dir_expanded'])
		checkbox = self.model.get_value(iter, self.model.column['checkbox'])
		self.model.expand(iter, True)
		return

	def set_prop(self, property):	
		# sets the property of fsh which determine types of browsed files
		if (property in self.model.dir_entries.pref.keys()):
			self.model.dir_entries.pref[property] = True

	def unset_prop(self, property):	
		# sets the property of fsh which determine types of browsed files
		if (property in self.model.dir_entries.pref.keys()):
			self.model.dir_entries.pref[property] = False


	def init(self, pathname, selected = False, expand = False):

		CView.init(self, [pathname], selected)
		iter, path_rest = self.model.get_iter_from_pathname(pathname)
		if selected:
			self.checked_path = self.model.get_path(iter)
		path = self.model.get_path(iter)
		self.set_cursor(path)

		# for wrapping dialog scrolling
		self.init_tv_path = path


################################################################
# CSelOnePathDialog
class CSelOnePathDialog(gtk.Dialog):
	"""
	This class creates modal dialog for path selection.

	methods:
	get_path -- run dialog, returns selected path or ''
	"""

	def __init__(self, aTitle='', aParent=None, message="Select path",
		     set_props=['hidden', 'empty', 'access', 'directory', 'sort'],
		     unset_props=['file', 'charDev', 'blockDev', 'fifo', 'symlink', 'socket']):		     
	        gtk.Dialog.__init__(self, title=aTitle, parent=aParent, flags=gtk.DIALOG_MODAL, buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                      gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
		gtk.Dialog.set_position(self, gtk.WIN_POS_CENTER_ON_PARENT)

		self.__message = message		    
		self.__label_msg = gtk.Label(self.__message)		    
		self.vbox.add(self.__label_msg)

		self.__fsh = COnePathSelection('/', False, set_props, unset_props)
		self.__scrollwin = gtk.ScrolledWindow()
		self.vbox.add(self.__scrollwin)
		old = self.vbox.query_child_packing(self.__label_msg)
		self.vbox.set_child_packing(self.__label_msg, 0, 0, 5, old[3])
		self.__scrollwin.add_with_viewport(self.__fsh)
		self.__ret_path = ""		

		self.set_size_request(400, 400)

		self.__running = False


	def get_path(self, pathname=''):
		"""
		Runs dialog and returns selected path or empty string.
		
		"""
		
		# wrapped gtk.Dialog.run method

		# dialog is modal, so this shouldn't occur
		if self.__running:
			return None
		self.__running = True

		self.__ret_path = ""
		
		home_dir = os.getenv("HOME")
		if pathname == '':
			sel = False
			pathname = home_dir
		else:
			sel = True
			
		self.__fsh.init(pathname, selected=sel)	
		self.connect('response', self.__response_cb)
		self.show_all()
			
		sel_path_y = self.__fsh.get_cell_area(self.__fsh.init_tv_path, self.__fsh.get_column(0)).y
		page_size = self.__scrollwin.get_vadjustment().page_size
		self.__scrollwin.get_vadjustment().set_value(max(sel_path_y - page_size/2, 0))
		
		# OUT: gtk.gdk.threads_enter()
		gtk.main()
		# OUT: gtk.gdk.threads_leave()

		self.__running = False
		self.destroy()

		return self.__ret_path
		
	def __response_cb(self, dialog = None, response = None, data1 = None):
		if response == gtk.RESPONSE_OK:
			self.__ret_path = self.__fsh.get_selected()
			if self.__ret_path == None or self.__ret_path == []:
				self.__ret_path = ""
			else:
				self.__ret_path = self.__ret_path[0]
	       	gtk.main_quit()

	def set_message(self, message):
		self.__message = message		    
		self.__label_msg.set_label(message)


	def set_fsh_prop(self, property):	
		# sets the property of fsh which determine types of browsed files
		self.__fsh.set_prop(property)

	def unset_fsh_prop(self, property):	
		# sets the property of fsh which determine types of browsed files
		self.__fsh.unset_prop(property)
		


################################################################
# CSelPathsDialog
class CSelPathDialog(gtk.Dialog):
	"""
	This class creates modal dialog for multiple paths selection.

	methods:
	get_paths -- run dialog, returns list of selected paths or []
	"""

	def __init__(self, aTitle='', aParent=None, message="Select path",
		     set_props=['hidden', 'empty', 'access', 'directory', 'file', 'sort'],
		     unset_props=['charDev', 'blockDev', 'fifo', 'symlink', 'socket']):		     
	        gtk.Dialog.__init__(self, title=aTitle, parent=aParent, flags=gtk.DIALOG_MODAL, buttons=(gtk.STOCK_OK, gtk.RESPONSE_OK,
                      gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
		gtk.Dialog.set_position(self, gtk.WIN_POS_CENTER_ON_PARENT)

		self.__message = message		    
		self.__label_msg = gtk.Label(self.__message)		    
		self.vbox.add(self.__label_msg)

		self.__fsh = CView('/', False, True)

		for prop in set_props:
			self.set_fsh_prop(prop)

		for prop in unset_props:
			self.unset_fsh_prop(prop)
		
		self.__scrollwin = gtk.ScrolledWindow()
		self.vbox.add(self.__scrollwin)
		old = self.vbox.query_child_packing(self.__label_msg)
		self.vbox.set_child_packing(self.__label_msg, 0, 0, 5, old[3])
		self.__scrollwin.add_with_viewport(self.__fsh)
		self.__ret_files = []		

		self.set_size_request(400, 400)

		self.__running = False


	def get_paths(self, paths=[], files_only=False):
		"""
		Runs dialog and returns selected path or ''.

		paths -- initial paths displayed as selected 
		files_only -- return only paths of displayed and checked files
		
		"""
		
		# wrapped gtk.Dialog.run method

		# dialog is modal, so this shouldn't occur
		if self.__running:
			return None
		self.__running = True

		home_dir = os.getenv("HOME")
		sel = True
		if paths == []:
			sel = False
			paths = [home_dir]

		self.__ret_paths = []

		self.__fsh.init(paths, selected=sel)	
		self.connect('response', self.__response_cb, files_only)
		self.show_all()

		if paths:
			first_path = paths[0]
		
			iter, path_rest = self.__fsh.model.get_iter_from_pathname(first_path)
			path = self.__fsh.model.get_path(iter)
			self.__fsh.set_cursor(path)
		
			sel_path_y = self.__fsh.get_cell_area(path, self.__fsh.get_column(0)).y
			page_size = self.__scrollwin.get_vadjustment().page_size
			self.__scrollwin.get_vadjustment().set_value(max(sel_path_y - page_size/2, 0))
		
		
		#gtk.gdk.threads_enter()
		gtk.main()
		#gtk.gdk.threads_leave()

		self.__running = False
		self.destroy()

		return self.__ret_paths
		
	def __response_cb(self, dialog = None, response=None, files_only=False):
		if response == gtk.RESPONSE_OK:
			if files_only:
				self.__ret_paths = self.__fsh.get_af_selected(['dir'])
			else:	
				self.__ret_paths = self.__fsh.get_selected()
			if self.__ret_paths == None:
				self.__ret_paths = []
	       	gtk.main_quit()

	def set_message(self, message):
		self.__message = message		    
		self.__label_msg.set_label(message)


	def set_fsh_prop(self, prop):	
		# sets the property of fsh which determine types of browsed files
		if (prop in self.__fsh.model.dir_entries.pref.keys()):
			self.__fsh.model.dir_entries.pref[prop] = True



	def unset_fsh_prop(self, prop):	
		# sets the property of fsh which determine types of browsed files
		if (prop in self.__fsh.model.dir_entries.pref.keys()):
			self.__fsh.model.dir_entries.pref[prop] = False
		

if __name__ == "__main__":

##	# CView
##	view = CView(["/home/qwertyn/"], True, True)
##	window = Window(WINDOW_TOPLEVEL)
##	window.add(view)
##	window.set_size_request(640,480)
##	window.show_all()
##	main()

        # for testing - gtk.Dialog.run() conflicts with gdk threads
        gtk.gdk.threads_init()

##      # CSelOnePathDialog	
##	sel_dlg = CSelOnePathDialog()
##	print sel_dlg.get_path('/var')
##	sel_dlg.set_title('New_title')
##	sel_dlg.set_message('Select some file path:')
##	sel_dlg.set_fsh_prop('file')
##	print sel_dlg.get_path()

	sel_dlg = CSelPathDialog()
	files = sel_dlg.get_paths([])
	sel_dlg = CSelPathDialog()
	sel_dlg.set_title('New_title')
	sel_dlg.set_message('Select files:')
	print sel_dlg.get_paths(files, files_only=True)

