"""
Module for getting various system information

Copyright (C) GRISOFT, s.r.o.
"""

import os
import sre
import pwd
import logger

import ConfigParser

################################################################
# CInfoBase
class CInfoBase:
	def __init__(self):
		self.values = {}

	def get_value(self, name):
		try:
			return self.values[name]
		except KeyError:
			raise AttributeError, name
			
	def set_value(self, name, value):
		try:
			self.values[name] = value
		except KeyError:
			raise AttributeError, name

################################################################
# CSysInfo
class CSysInfo(CInfoBase):
	"""
	CSysInfo is class that provides several system information
	"""
	
	def __init__(self):
		CInfoBase.__init__(self)

		self.values = {
			'user_name'	: "N/A",
			'user_id'	: "N/A",

			'sys_name'	: "N/A",
			'hostname'	: "N/A",
			'release'	: "N/A",
			'version'	: "N/A",
			'machine'	: "N/A"
		}

		try:
			uname = os.uname()

			self.values['user_name'] = pwd.getpwuid(os.geteuid()).pw_name
			self.values['user_id']   = os.geteuid()
			self.values['sys_name']  = uname[0]
			self.values['hostname']  = uname[1]
			self.values['release']   = uname[2]
			self.values['version']   = uname[3]
			self.values['machine']   = uname[4]
		except OSError:
			pass

################################################################
# CAvgInfo
def CAvgInfo():
	"""CAvgInfo is a singleton. This function returns handle to it."""

	if __CAvgInfo.instance == None:
		__CAvgInfo.instance = __CAvgInfo()

	return __CAvgInfo.instance

class __CAvgInfo(CInfoBase):

	instance = None

	def __init__(self):
		CInfoBase.__init__(self)

		self.values = {
			'avgver'	: ("", []),
			'aviver'	: ("", []),
			'copyright'	: (" Copyright (c) 2006 GRISOFT, Inc.", []),
			'user_name'	: ("N/A", []),
			'company'	: ("N/A", []),
			'lic_num'	: ("N/A", []),
			'lic_type'      : ("", [])
		}

		# NOTE: At this moment, only avgver, aviver and lic_type are kept up-to-date.

		self.avgscan_output_list = ""

		self.__logger = logger.CLogger()

		# get user name and company from config
		config = ConfigParser.ConfigParser()
		try:
			config.readfp(open('/opt/grisoft/avggui/config/userinfo.cfg'))

			self.values['user_name'] = (config.get('license_user', 'username'), [])
			self.values['company']   = (config.get('license_user', 'company'), [])
			self.values['lic_num']   = (config.get('license_user', 'license_number'), [])

		except IOError, filename:
			self.__logger.err("opening file: %s" % filename)
		except ConfigParser.NoSectionError, sect_name:
			pass
		except ConfigParser.NoOptionError, option:
			pass

		self.update()

	def update(self, avgscan_output_list = None):	
		"""Updates aviver, agvvel and lic_type according to avgscan output."""

		# NOTE: only when value has changed, set_value method is used and thus
		# callbacks are called.
		
		if avgscan_output_list == None:
			avgscan_output      = os.popen("avgscan 2>&1", "r", 1)
			self.avgscan_output_list   = avgscan_output.readlines()
		else:
			self.avgscan_output_list   = avgscan_output_list

		# get license number, avi version, avg version from avgscan
		(avgver, aviver) = self.__get_avgavi_ver(self.avgscan_output_list)
		lic_type = self.__get_license_type(self.avgscan_output_list)

		if self.get_value('avgver') != avgver:
			self.set_value('avgver', avgver)
		if self.get_value('aviver') != aviver:
			self.set_value('aviver', aviver)
		if self.get_value('lic_type') != lic_type:
			self.set_value('lic_type', lic_type) 


	def set_value(self, name = None, value = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		self.values[name] = (value, cbs)
		
		# call callbacks with new value
		for cb in cbs:
			cb(value)
		
		self.__logger.debug("Setting value '%s' to name '%s'." % (value, name))

	def get_value(self, name = None):
		try:
			(v, c) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		return v
		
	def connect(self, name = None, cb = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
			
		cbs.append(cb)
		self.values[name] = (v, cbs)
	
	def disconnect(self, name = None, cb = None):
		try:
			(v, cbs) = self.values[name]
		except KeyError:
			raise AttributeError, name
		
		cbs.remove(cb)
		self.values[name] = (v, cbs)

	def __get_avgavi_ver(self, avgscan_output_list):
		avg_ver, avi_ver = "", ""
		
		for line in avgscan_output_list:
			if sre.match("^Program version", line):
				avg_ver = sre.split(" ", line)[2][:-1]
			elif sre.match("^Virus Database", line):
				avi_ver = sre.split(" ", line)[3]
		
		return (avg_ver, avi_ver)
	

		
	def __get_license_type(self, avgscan_output_list):
		"""Get license type of instaled AVG (FREE| FULL| TRIAL| SALES)"""

		# NOTE: License ERROR case (expired) shouldn't occur (avggui wouldn't have been started in the case)

		for line in avgscan_output_list:
			if sre.match("^License type", line):
				try:
					return sre.search ("(FREE|TRIAL|SALES|FULL)", line).group()
				except:
					self.__logger.err("Can't determine type of license from line \"" + line + "\"")
					return ""
					
################################################################
# testing code
if __name__ == "__main__":
	
	sysinfo = CAvgInfo()
	
	print sysinfo.get_value("lic_num")
