"""
Module for manipulating with test results from avgscan.

Copyright (C) GRISOFT, s.r.o.
"""

import config
import logger

import os
import sre
import time
import types

################################################################
# CTestResultFile
class CTestResultFile:
	"""
	CTestResultFile is class for reading results from output result file
	of avgscan.
	It provides iterator so it could be used in for cycle.
	Example: for (tested_file, result) in CTestResultFile(file):
	             print tested_file, result
	
	Attributes:
		file -- result file specified by file object or by path string
		line -- current line in the file
		
		scanned_files    -- number of scanned files
		scanned_sectors  -- number of scanned sectors
		infected_files   -- number of infeted files
		infected_sectors -- number of infected sectors
		errors           -- number of errors
		reported_files   -- number of reported files
	
	Methods:
		next -- for iterating in file
		get_stats -- get statistics from file
	"""
	
	def __init__(self, file):
		if type(file) is types.StringType:
			self.file = open(file, "rw")
		elif type(file) is types.FileType:
			self.file = file
		else:
			raise IOError

		# Position in results file (relative the end of file)
		# from which we can look for statistics. 
		self.__stats_position = -6*60
		
		# skip program info text
		self.line = self.file.readline()
		while self.line and sre.match("^\".*\"", self.line) == None and sre.match("^Test start", self.line) == None:
			self.line = self.file.readline()
		
		self.scanned_files    = 0
		self.scanned_sectors  = 0
		self.infected_files   = 0
		self.infected_sectors = 0
		self.errors = 0
		self.reported_files = None

	def fill_stats(self):
		"""fill statistics attributes from file (position of iterator is preserved)."""		

		saved_position = self.file.tell()
		self.file.seek(self.__stats_position,2)
		self.__readStats()
		self.file.seek(saved_position)				
		
	def __readStats(self):
		while self.line:
			num = sre.split("\s*:\s*", self.line)
			
			if len(num) > 1:
				try:
					num = int(num[1])
				except ValueError:	
					pass
			
			if sre.match("^Scanned.*file", self.line):
				self.scanned_files = num
			elif sre.match("^Scanned.*sectors", self.line):
				self.scanned_sectors = num
			elif sre.match("^Infected.*file", self.line):
				self.infected_files = num
			elif sre.match("^Infected.*sector", self.line):
				self.infected_sectors = num
			elif sre.match("^Errors", self.line):
				self.errors = num
			elif sre.match("^Reported.*file", self.line):
				self.reported_files = num	
				
			self.line = self.file.readline()
			
	def __iter__(self):
		return self
	
	def next(self):
		if self.line:
			l_parts = sre.split("\"", sre.sub("^\"", "", self.line))
			
			if len(l_parts) < 2:
				self.__readStats()
				raise StopIteration
			
			l_parts[1] = sre.sub("^[\s]*", "", l_parts[1])
			l_parts[1] = l_parts[1][:-1]
		else:
			raise StopIteration
		
		self.line = self.file.readline()
		
		return (l_parts[0], l_parts[1])

	def close(self):
		self.file.close()

################################################################
# CTestResultsList
class CTestResultsList:
	"""
	CTestResultsList is class for reading results from output result file
	of avgscan.
	
	Attributes:
	
	Methods:
		results_list -- list of tupples (filename, timestamp, bool).
		                The boolean item is False if all tested files
						are clean.
	"""

	def __init__(self):
		self.conf = config.CConfigStorage
		self.__logger = logger.CLogger()
		
		results_path = self.conf.test_res_config.get_value("res_path")

		if os.access(results_path, os.R_OK | os.X_OK) == 0:
			raise IOError, results_path

		self.results_list = []

	def __delete_overfull(self):
		try:
			max_files = self.conf.test_res_config.get_value("num_saved")

			results_path = self.conf.test_res_config.get_value("res_path")

			files = os.listdir(results_path)
			
			for f in files:
				if os.path.isdir(f):
					files.remove(f)
					
			files.sort()
		except AttributeError, which:
			self.__logger.err("AttributeError: %s" % which)
			files = None

			return
		except OSError, details:
			self.__logger.err("OSError: %s" % details)
			return
		
		diff_len  = len(files) - max_files
		
		if diff_len > 0:
			del_files = files[:diff_len]
		else:
			del_files = []
		
		for f in del_files:
			os.unlink("%s/%s" % (results_path, f))

	def __delete_old(self):
		act_time = time.time()
		try:
			secs_in_maxdays = self.conf.test_res_config.get_value("del_older") * 24 * 60 * 60

			results_path = self.conf.test_res_config.get_value("res_path")

			files = os.listdir(results_path)
			
			for f in files:
				if os.path.isdir(f):
					files.remove(f)

			files.sort()
		except AttributeError, which:
			self.__logger.err("AttributeError: %s" % which)
			files = None
			
			return
		except OSError, details:
			self.__logger.err("OSError: %s" % details)
			return

		limit_timestamp = int(act_time - secs_in_maxdays)

		for f in files:
			if sre.match("testreport\..*", f):
				mtime = int(f.split(".")[1])

				if mtime < limit_timestamp:
					os.unlink("%s/%s" % (results_path, f))
				
	def __process_file(self, filename):
		
		filename_parts = filename.split(".")
		
		return (filename, filename_parts[1], len(filename_parts) != 2)
		
	def get_results_list(self):
		"""
		Return list of tupples with filename, timestamp and value of test
		result.
		"""
		
		return self.results_list
	
	def update(self):
		"""
		Actualize list of stored files.
		"""
		
		self.results_list = []
		
		try:
			results_path = self.conf.test_res_config.get_value("res_path")
		except AttributeError:
			return
			
		
		self.__delete_overfull()
		self.__delete_old()

		files = os.listdir(results_path)
		files.sort()

		for f in files:
			if sre.match("testreport\..*", f):
				self.results_list.append(self.__process_file(f))
		

################################################################
# CTestResultsList
class CTestResultsInDirList:
	"""
	CTestResultsListInDir is class for reading results from output result file
	of avgscan.
	
	Parameters:
		results_path -- path to result files

	Attributes:
		results_list -- list of tupples (filename, timestamp, bool).
		                The boolean item is False if all tested files
						are clean.
	Methods:
		get_results_list(number=None, since=None) 	-- returns result_list
		update 				-- rereads informaton from disk

	"""

	def __init__(self, results_path):

		self.__logger = logger.CLogger()

		self.__results_path = results_path

		if os.access(self.__results_path, os.R_OK | os.X_OK) == 0:
			raise IOError, self.__results_path

		self.results_list = []

	def __process_file(self, filename):
		
		filename_parts = filename.split(".")		
		return (filename, filename_parts[1], len(filename_parts) != 2)
		
	def get_results_list(self, number=None, min_timestamp=None):
		"""
		Return list of tupples with filename, timestamp and value of test
		result. Only number of most recent results (given by arg. number)
		and/or only results newer than min_timestamp (in seconds) 
		can be returned. List is sorted with increasing timestamp.

		"""

		filtered_list = self.results_list

		if (number > 0):	
			filtered_list[:-number] = []
		elif (number == 0) :	
			filtered_list = []

		if (min_timestamp > 0):		
			filtered_list = [result for result in filtered_list if int(result[1]) > int(min_timestamp)]   
		
		return filtered_list
	
	def update(self):
		"""
		Actualize list of stored files.
		Raises OSError if source path is no more accessible. 
		"""
		
		self.results_list = []
		
		if os.access(self.__results_path, os.R_OK | os.X_OK) == 0:
			raise IOError, self.__results_path
		
		try:
			files = os.listdir(self.__results_path)		
		except OSError, which:
			self.__logger.err("OSError: %s" % which)
			files = None	
		
		self.results_list = [self.__process_file(f) for f in files if sre.match("testreport\..*", f)]		
		self.results_list.sort(lambda l1,l2: int(l1[1])-int(l2[1]))									




################################################################
# testing code
if __name__ == "__main__":

	try:
		res = CTestResultFile("testreport.txt")
		
		for (n, r) in res:
			print n, " ** ", r

		print res.scanned_files
		print res.infected_files
		
		conf = config.CConfigStorage
		conf.test_res_config.set_value("res_path", "testresults")
		conf.test_res_config.set_value("num_saved", 5)
		
		res_list = CTestResultsList()
		res_list.update()
		res_list.get_results_list()
		
	except IOError, what:
		print "** ERR: File IOError: %s" % what
	except AttributeError, which:
		print "** ERR: Attribute Error: %s" % which
		
